﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;

namespace ARSampleApp
{
	public partial class MainPage : PhoneApplicationPage
	{
		// Constructor
		public MainPage()
		{
			InitializeComponent();
			//Add some more elements to the panel besides the ones already added in XAML
			AddHeadingDots(arPanel);
		}

		private void ARPanel_Loaded(object sender, RoutedEventArgs e)
		{
			var panel = sender as SharpGIS.AR.Controls.ARPanel;
			//For motion to be supported, device needs at least a compass and
			//an accelerometer. A gyro as well makes the experience much better though
			if (Microsoft.Devices.Sensors.Motion.IsSupported)
			{
				//Warn user that without gyro, the experience isn't as good as it can get
				if (!Microsoft.Devices.Sensors.Gyroscope.IsSupported)
				{
					LayoutRoot.Children.Add(new TextBlock()
					{
						Text = "No gyro detected. Experience may be degraded",
						TextWrapping = System.Windows.TextWrapping.Wrap,
						VerticalAlignment = System.Windows.VerticalAlignment.Bottom
					});
				}
				//Start the AR PAnel
				panel.Start();
			}
			else //Bummer! 
			{
				panel.Visibility = System.Windows.Visibility.Collapsed;
				MessageBox.Show("Sorry - Motion sensor is not supported on this device");
			}
		}

		private void ARPanel_Unloaded(object sender, RoutedEventArgs e)
		{
			//Remember to stop the motion sensor when leaving
			(sender as SharpGIS.AR.Controls.ARPanel).Stop();
		}

		private void AddHeadingDots(SharpGIS.AR.Controls.ARPanel panel)
		{
			//Go 360 along the horizon to display heading text
			for (double heading = 20; heading < 360; heading += 20)
			{
				if (heading % 90 == 0) continue; //skip cardinal directions, since we already added markers for those in xaml
				TextBlock tb = new TextBlock() { Text = string.Format("{0}°", heading) };
				SharpGIS.AR.Controls.ARPanel.SetDirection(tb, new Point(0, heading));
				panel.Children.Add(tb);
			}
			//Display an up/down angle for each cardinal direction
			for (int i = 0; i < 360; i += 90)
			{
				for (int azimuth = -80; azimuth < 90; azimuth += 10)
				{
					if (azimuth == 0) continue; //skip cardinal directions, since we already added markers for those in xaml
					TextBlock tb = new TextBlock() { Text = string.Format("{0}", azimuth) };
					SharpGIS.AR.Controls.ARPanel.SetDirection(tb, new Point(azimuth, i));
					panel.Children.Add(tb);
				}
			}
		}
	}
}