//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft
// premium shared source license agreement under which you licensed
// this source code. If you did not accept the terms of the license
// agreement, you are not authorized to use this source code.
// For the terms of the license, please see the license agreement
// signed by you and Microsoft.
// THE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//
//@include "..\Preprocessed.h"

using System;
using System.Globalization;
using System.Collections.Generic;
using System.Text;
using SM=System.Math;
using System.Diagnostics.CodeAnalysis;

namespace Microsoft.Xna.Framework
{
    public struct Matrix : IEquatable<Matrix>
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M11;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M12;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M13;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M14;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M21;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M22;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M23;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M24;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M31;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M32;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M33;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M34;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M41;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M42;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M43;
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public float M44;

        #region Static Properties
        private static Matrix _identity = new Matrix(
            1.0f, 0.0f, 0.0f, 0.0f,
            0.0f, 1.0f, 0.0f, 0.0f,
            0.0f, 0.0f, 1.0f, 0.0f,
            0.0f, 0.0f, 0.0f, 1.0f );
 
        public static Matrix Identity { get { return _identity; } }
        #endregion

        #region Properties
        /// <summary>
        /// Gets, Sets Up vector of matrix.
        /// </summary>
        public Vector3 Up
        {
            get
            {
                Vector3 ans;
                ans.X = M21; ans.Y = M22; ans.Z = M23;
                return ans;
            }
            set
            {
                M21 = value.X; M22 = value.Y; M23 = value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Down vector of matrix.
        /// </summary>
        public Vector3 Down
        {
            get
            {
                Vector3 ans;
                ans.X = -M21; ans.Y = -M22; ans.Z = -M23;
                return ans;
            }
            set
            {
                M21 = -value.X; M22 = -value.Y; M23 = -value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Right vector of matrix.
        /// </summary>
        public Vector3 Right
        {
            get
            {
                Vector3 ans;
                ans.X = M11; ans.Y = M12; ans.Z = M13;
                return ans;
            }
            set
            {
                M11 = value.X; M12 = value.Y; M13 = value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Left vector of matrix.
        /// </summary>
        public Vector3 Left
        {
            get
            {
                Vector3 ans;
                ans.X = -M11; ans.Y = -M12; ans.Z = -M13;
                return ans;
            }
            set
            {
                M11 = -value.X; M12 = -value.Y; M13 = -value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Forward vector of matrix.
        /// </summary>
        public Vector3 Forward
        {
            get
            {
                Vector3 ans;
                ans.X = -M31; ans.Y = -M32; ans.Z = -M33;
                return ans;
            }
            set
            {
                M31 = -value.X; M32 = -value.Y; M33 = -value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Backward vector of matrix.
        /// </summary>
        public Vector3 Backward
        {
            get
            {
                Vector3 ans;
                ans.X = M31; ans.Y = M32; ans.Z = M33;
                return ans;
            }
            set
            {
                M31 = value.X; M32 = value.Y; M33 = value.Z;
            }
        }

        /// <summary>
        /// Gets, Sets Translation part of matrix.
        /// </summary>
        public Vector3 Translation
        {
            get
            {
                Vector3 ans;
                ans.X = M41; ans.Y = M42; ans.Z = M43;
                return ans;
            }
            set
            {
                M41 = value.X; M42 = value.Y; M43 = value.Z;
            }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Construct Matrix by given element values.
        /// </summary>
        [SuppressMessage("Microsoft.Design","CA1025")]//suppressing fxcop's suggestion to make a ctor taking a variable number of arguments
        public Matrix(
            float m11, float m12, float m13, float m14,
            float m21, float m22, float m23, float m24,
            float m31, float m32, float m33, float m34,
            float m41, float m42, float m43, float m44 )
        {
            M11 = m11; M12 = m12; M13 = m13; M14 = m14;
            M21 = m21; M22 = m22; M23 = m23; M24 = m24;
            M31 = m31; M32 = m32; M33 = m33; M34 = m34;
            M41 = m41; M42 = m42; M43 = m43; M44 = m44;
        }
        #endregion

        #region Factory methods
        /// <summary>
        /// Create spherical billboard that rotates around specfied object position.
        /// Vector3.Forward vector of local coordinate facing towards to camera position.
        /// Vector3.Up vector of local coordinate aligned to cameraUpVector.
        /// </summary>
        /// <param name="objectPosition"></param>
        /// <param name="cameraPosition"></param>
        /// <param name="cameraUpVector">mut be unit vector</param>
        /// <param name="cameraForwardVector">Optional: specify camera foward vector. unit vector.</param>
        /// <returns>Created billboard matrix.</returns>
        /// <remarks>This method computes facing direction from object position and camera position.
        /// When object and camera positions are too close, it can't compute proper matrix.
        /// To avoid this problem, this method uses Vector3.Forward vector when that situation.
        /// This assumption is good as long as camera forward direction is close to Vector3.Foward.
        /// Otherwise you see the 'pop' effect in this situation.
        /// You can reduce this issue by specify cameraForwardVector
        /// </remarks>
        public static Matrix CreateBillboard(Vector3 objectPosition, Vector3 cameraPosition, Vector3 cameraUpVector, Vector3? cameraForwardVector)
        {
            const float epsilon = 1e-4f;

            Vector3 zaxis;
            zaxis.X = objectPosition.X - cameraPosition.X;
            zaxis.Y = objectPosition.Y - cameraPosition.Y;
            zaxis.Z = objectPosition.Z - cameraPosition.Z;
            float norm = zaxis.LengthSquared();
            if (norm < epsilon)
            {
                zaxis = cameraForwardVector.HasValue ? -cameraForwardVector.Value : Vector3.Forward;
            }
            else
            {
                Vector3.Multiply(ref zaxis, 1.0f / (float)Math.Sqrt(norm), out zaxis);
            }

            Vector3 xaxis;
            Vector3.Cross(ref cameraUpVector, ref zaxis, out xaxis);
            xaxis.Normalize();

            Vector3 yaxis;
            Vector3.Cross(ref zaxis, ref xaxis, out yaxis);

            Matrix result;
            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;

            result.M41 = objectPosition.X;
            result.M42 = objectPosition.Y;
            result.M43 = objectPosition.Z;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateBillboard(ref Vector3 objectPosition, ref Vector3 cameraPosition, ref Vector3 cameraUpVector, Vector3? cameraForwardVector, out Matrix result)
        {
            const float epsilon = 1e-4f;

            Vector3 zaxis;
            zaxis.X = objectPosition.X - cameraPosition.X;
            zaxis.Y = objectPosition.Y - cameraPosition.Y;
            zaxis.Z = objectPosition.Z - cameraPosition.Z;
            float norm = zaxis.LengthSquared();
            if (norm < epsilon)
            {
                zaxis = cameraForwardVector.HasValue ? -cameraForwardVector.Value : Vector3.Forward;
            }
            else
            {
                Vector3.Multiply(ref zaxis, 1.0f / (float)Math.Sqrt(norm), out zaxis);
            }

            Vector3 xaxis;
            Vector3.Cross(ref cameraUpVector, ref zaxis, out xaxis);
            xaxis.Normalize();

            Vector3 yaxis;
            Vector3.Cross(ref zaxis, ref xaxis, out yaxis);

            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;

            result.M41 = objectPosition.X;
            result.M42 = objectPosition.Y;
            result.M43 = objectPosition.Z;
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Create cylindalical billboard that rotating around specfied axis.
        /// rotateAxis matches y-axis of local coordinate and Vector3.Forward of object coordinate facing towards to camera position.
        /// </summary>
        /// <param name="objectPosition"></param>
        /// <param name="cameraPosition"></param>
        /// <param name="rotateAxis">must be unit vector.</param>
        /// <param name="cameraForwardVector">Optional: specify camera foward vector. unit vector.</param>
        /// <param name="objectForwardVector">Optional: specify object foward vector. unit vector.</param>
        /// <returns>Created billboard matrix.</returns>
        /// <remarks>This method computes facing direction from object position and camera position.
        /// When object and camera positions are too close, it can't compute proper matrix.
        /// To avoid this problem, this method uses Vector3.Forward vector when that situation.
        /// This assumption is good as long as camera forward direction is close to Vector3.Foward.
        /// Otherwise you see the 'pop' effect in this situation.
        /// You can reduce this issue by specify cameraForwardVector
        /// 
        /// Also when angle between camera to object direction and rotateAxis become close to 0, it cant compute proper matrix either.
        /// To avoid this problem, this method tries to use Vector3.Forward then Vector3.Right vector for object facing direction.
        /// You can change this behavior by passing objectForwardVector.
        /// Usually you want to pass an object forward direction from previous frame computation result.
        /// So it keeps previous frame object rotation while this problem occurring.
        /// </remarks>
        public static Matrix CreateConstrainedBillboard(Vector3 objectPosition, Vector3 cameraPosition, Vector3 rotateAxis,
                        Vector3? cameraForwardVector, Vector3? objectForwardVector )
        {
            const float epsilon = 1e-4f;
            const float minAngle = 1.0f - (0.1f * (MathHelper.Pi / 180.0f)); // 0.1 degrees

            // Treat the case when object and camera positions are too close.
            Vector3 faceDir;
            faceDir.X = objectPosition.X - cameraPosition.X;
            faceDir.Y = objectPosition.Y - cameraPosition.Y;
            faceDir.Z = objectPosition.Z - cameraPosition.Z;

            float norm = faceDir.LengthSquared();
            if (norm < epsilon)
            {
                faceDir = cameraForwardVector.HasValue ? -cameraForwardVector.Value : Vector3.Forward;
            }
            else
            {
                Vector3.Multiply(ref faceDir, (1.0f / (float)Math.Sqrt(norm)), out faceDir);
            }

            Vector3 yaxis = rotateAxis;
            Vector3 xaxis;
            Vector3 zaxis;

            // Treat the case when angle between faceDir and rotateAxis is too close to 0.
            float dot;
            Vector3.Dot(ref rotateAxis, ref faceDir, out dot);
            if (SM.Abs(dot) > minAngle)
            {
                if (objectForwardVector.HasValue)
                {
                    zaxis = objectForwardVector.Value;
                    // Make sure passed values are useful for compute.
                    Vector3.Dot(ref rotateAxis, ref zaxis, out dot);
                    if (SM.Abs(dot) > minAngle)
                    {
                        dot = rotateAxis.X * Vector3.Forward.X + rotateAxis.Y * Vector3.Forward.Y + rotateAxis.Z * Vector3.Forward.Z;
                        zaxis = (SM.Abs(dot) > minAngle) ? Vector3.Right : Vector3.Forward;
                    }
                }
                else
                {
                    dot = rotateAxis.X * Vector3.Forward.X + rotateAxis.Y * Vector3.Forward.Y + rotateAxis.Z * Vector3.Forward.Z;
                    zaxis = (SM.Abs(dot) > minAngle) ? Vector3.Right : Vector3.Forward;
                }

                Vector3.Cross(ref rotateAxis, ref zaxis, out xaxis);
                xaxis.Normalize();

                Vector3.Cross(ref xaxis, ref rotateAxis, out zaxis);
                zaxis.Normalize();
            }
            else
            {
                Vector3.Cross(ref rotateAxis, ref faceDir, out xaxis);
                xaxis.Normalize();

                Vector3.Cross(ref xaxis, ref yaxis, out zaxis);
                zaxis.Normalize();
            }

            Matrix result;
            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;

            result.M41 = objectPosition.X;
            result.M42 = objectPosition.Y;
            result.M43 = objectPosition.Z;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateConstrainedBillboard(ref Vector3 objectPosition, ref Vector3 cameraPosition, ref Vector3 rotateAxis,
                        Vector3? cameraForwardVector, Vector3? objectForwardVector, out Matrix result)
        {
            const float epsilon = 1e-4f;
            const float minAngle = 1.0f - (0.1f * (MathHelper.Pi / 180.0f)); // 0.1 degrees

            // Treat the case when object and camera positions are too close.
            Vector3 faceDir;
            faceDir.X = objectPosition.X - cameraPosition.X;
            faceDir.Y = objectPosition.Y - cameraPosition.Y;
            faceDir.Z = objectPosition.Z - cameraPosition.Z;

            float norm = faceDir.LengthSquared();
            if (norm < epsilon)
            {
                faceDir = cameraForwardVector.HasValue ? -cameraForwardVector.Value : Vector3.Forward;
            }
            else
            {
                Vector3.Multiply(ref faceDir, (1.0f / (float)Math.Sqrt(norm)), out faceDir);
            }

            Vector3 yaxis = rotateAxis;
            Vector3 xaxis;
            Vector3 zaxis;

            // Treat the case when angle between faceDir and rotateAxis is too close to 0.
            float dot;
            Vector3.Dot(ref rotateAxis, ref faceDir, out dot);
            if (SM.Abs(dot) > minAngle)
            {
                if (objectForwardVector.HasValue)
                {
                    zaxis = objectForwardVector.Value;
                    // Make sure passed values are useful for compute.
                    Vector3.Dot(ref rotateAxis, ref zaxis, out dot);
                    if (SM.Abs(dot) > minAngle)
                    {
                        dot = rotateAxis.X * Vector3.Forward.X + rotateAxis.Y * Vector3.Forward.Y + rotateAxis.Z * Vector3.Forward.Z;
                        zaxis = (SM.Abs(dot) > minAngle) ? Vector3.Right : Vector3.Forward;
                    }
                }
                else
                {
                    dot = rotateAxis.X * Vector3.Forward.X + rotateAxis.Y * Vector3.Forward.Y + rotateAxis.Z * Vector3.Forward.Z;
                    zaxis = (SM.Abs(dot) > minAngle) ? Vector3.Right : Vector3.Forward;
                }

                Vector3.Cross(ref rotateAxis, ref zaxis, out xaxis);
                xaxis.Normalize();

                Vector3.Cross(ref xaxis, ref rotateAxis, out zaxis);
                zaxis.Normalize();
            }
            else
            {
                Vector3.Cross(ref rotateAxis, ref faceDir, out xaxis);
                xaxis.Normalize();

                Vector3.Cross(ref xaxis, ref yaxis, out zaxis);
                zaxis.Normalize();
            }

            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;

            result.M41 = objectPosition.X;
            result.M42 = objectPosition.Y;
            result.M43 = objectPosition.Z;
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Create translation matrix by given position.
        /// </summary>
        /// <param name="position"></param>
        /// <returns></returns>
        public static Matrix CreateTranslation(Vector3 position)
        {
            Matrix result;

            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;

            result.M41 = position.X;
            result.M42 = position.Y;
            result.M43 = position.Z;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateTranslation(ref Vector3 position, out Matrix result)
        {
            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;

            result.M41 = position.X;
            result.M42 = position.Y;
            result.M43 = position.Z;
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Create translation matrix by given position.
        /// </summary>
        /// <param name="xPosition"></param>
        /// <param name="yPosition"></param>
        /// <param name="zPosition"></param>
        /// <returns></returns>
        public static Matrix CreateTranslation( float xPosition, float yPosition, float zPosition )
        {
            Matrix result;

            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;

            result.M41 = xPosition;
            result.M42 = yPosition;
            result.M43 = zPosition;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateTranslation(float xPosition, float yPosition, float zPosition, out Matrix result)
        {
            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;

            result.M41 = xPosition;
            result.M42 = yPosition;
            result.M43 = zPosition;
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Create scale matrix by given parameters.
        /// </summary>
        /// <param name="xScale"></param>
        /// <param name="yScale"></param>
        /// <param name="zScale"></param>
        /// <returns></returns>
        public static Matrix CreateScale( float xScale, float yScale, float zScale )
        {
            float xs = xScale;
            float ys = yScale;
            float zs = zScale;

            Matrix result;
            result.M11 =   xs; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ys; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   zs; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;

            return result;
        }

        public static void CreateScale(float xScale, float yScale, float zScale, out Matrix result)
        {
            float xs = xScale;
            float ys = yScale;
            float zs = zScale;

            result.M11 =   xs; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ys; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   zs; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// Create scale matrix by given vector3.
        /// </summary>
        /// <param name="scales">scale vector that contains x, y, and z scales.</param>
        /// <returns></returns>
        public static Matrix CreateScale( Vector3 scales )
        {
            float xs = scales.X;
            float ys = scales.Y;
            float zs = scales.Z;

            Matrix result;
            result.M11 =   xs; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ys; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   zs; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
            return result;
        }

        public static void CreateScale(ref Vector3 scales, out Matrix result)
        {
            float xs = scales.X;
            float ys = scales.Y;
            float zs = scales.Z;

            result.M11 =   xs; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ys; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   zs; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// Create uniform scale matrix
        /// </summary>
        /// <param name="scale"></param>
        /// <returns></returns>
        public static Matrix CreateScale( float scale )
        {
            Matrix result;

            float ss = scale;
            result.M11 =   ss; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ss; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   ss; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;

            return result;
        }

        public static void CreateScale(float scale, out Matrix result)
        {
            float ss = scale;
            result.M11 =   ss; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =   ss; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 =   ss; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// Returns X-axis rotate matrix
        /// </summary>
        /// <param name="radians"></param>
        /// <returns></returns>
        public static Matrix CreateRotationX(float radians)
        {
            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  1  0  0  0 ]
            // [  0  c  s  0 ]
            // [  0 -s  c  0 ]
            // [  0  0  0  1 ]
            Matrix result;
            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =    c; result.M23 =    s; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 =   -s; result.M33 =    c; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;

            return result;
        }

        public static void CreateRotationX(float radians, out Matrix result)
        {
            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  1  0  0  0 ]
            // [  0  c  s  0 ]
            // [  0 -s  c  0 ]
            // [  0  0  0  1 ]
            result.M11 = 1.0f; result.M12 = 0.0f; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 =    c; result.M23 =    s; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 =   -s; result.M33 =    c; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// Returns Y-axis rotate matrix
        /// </summary>
        /// <param name="radians"></param>
        /// <returns></returns>
        public static Matrix CreateRotationY(float radians)
        {
            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  c  0 -s  0 ]
            // [  0  1  0  0 ]
            // [  s  0  c  0 ]
            // [  0  0  0  1 ]
            Matrix result;

            result.M11 =    c; result.M12 = 0.0f; result.M13 =   -s; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 =    s; result.M32 = 0.0f; result.M33 =    c; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;

            return result;
        }

        public static void CreateRotationY(float radians, out Matrix result)
        {
            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  c  0 -s  0 ]
            // [  0  1  0  0 ]
            // [  s  0  c  0 ]
            // [  0  0  0  1 ]
            result.M11 =    c; result.M12 = 0.0f; result.M13 =   -s; result.M14 = 0.0f;
            result.M21 = 0.0f; result.M22 = 1.0f; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 =    s; result.M32 = 0.0f; result.M33 =    c; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// Returns Y-axis rotate matrix
        /// </summary>
        /// <param name="radians"></param>
        /// <returns></returns>
        public static Matrix CreateRotationZ(float radians)
        {
            Matrix result;

            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  c  s  0  0 ]
            // [ -s  c  0  0 ]
            // [  0  0  1  0 ]
            // [  0  0  0  1 ]
            result.M11 =    c; result.M12 =    s; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 =   -s; result.M22 =    c; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;

            return result;
        }

        public static void CreateRotationZ(float radians, out Matrix result)
        {
            float c = (float)SM.Cos(radians);
            float s = (float)SM.Sin(radians);

            // [  c  s  0  0 ]
            // [ -s  c  0  0 ]
            // [  0  0  1  0 ]
            // [  0  0  0  1 ]
            result.M11 =    c; result.M12 =    s; result.M13 = 0.0f; result.M14 = 0.0f;
            result.M21 =   -s; result.M22 =    c; result.M23 = 0.0f; result.M24 = 0.0f;
            result.M31 = 0.0f; result.M32 = 0.0f; result.M33 = 1.0f; result.M34 = 0.0f;
            result.M41 = 0.0f; result.M42 = 0.0f; result.M43 = 0.0f; result.M44 = 1.0f;
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="axis">unit vector.</param>
        /// <param name="angle">in radians.</param>
        /// <returns></returns>
        public static Matrix CreateFromAxisAngle( Vector3 axis, float angle )
        {
            // a: angle
            // x, y, z: unit vector for axis.
            //
            // Rotation matrix M can compute by using below equation.
            //
            //        T               T
            //  M = uu + (cos a)( I-uu ) + (sin a)S
            //
            // Where:
            //
            //  u = ( x, y, z )
            //
            //      [  0 -z  y ]
            //  S = [  z  0 -x ]
            //      [ -y  x  0 ]
            //
            //      [ 1 0 0 ]
            //  I = [ 0 1 0 ]
            //      [ 0 0 1 ]
            //
            //
            //     [  xx+cosa*(1-xx)   yx-cosa*yx-sina*z zx-cosa*xz+sina*y ]
            // M = [ xy-cosa*yx+sina*z    yy+cosa(1-yy)  yz-cosa*yz-sina*x ]
            //     [ zx-cosa*zx-sina*y zy-cosa*zy+sina*x   zz+cosa*(1-zz)  ]
            //
            float x = axis.X, y = axis.Y, z = axis.Z;
            float sa = (float)Math.Sin(angle), ca = (float)Math.Cos(angle);
            float xx = x * x, yy = y * y, zz = z * z;
            float xy = x * y, xz = x * z, yz = y * z;

            Matrix result;

            result.M11 = xx + ca * ( 1.0f - xx );   result.M12 = xy - ca * xy + sa * z;     result.M13 = xz - ca * xz - sa * y;     result.M14 = 0.0f;
            result.M21 = xy - ca * xy - sa * z;     result.M22 = yy + ca * ( 1.0f - yy );   result.M23 = yz - ca * yz + sa * x;     result.M24 = 0.0f;
            result.M31 = xz - ca * xz + sa * y;     result.M32 = yz - ca * yz - sa * x;     result.M33 = zz + ca * ( 1.0f - zz );   result.M34 = 0.0f;
            result.M41 = 0.0f;                      result.M42 = 0.0f;                      result.M43 = 0.0f;                      result.M44 = 1.0f;

            return result;
        }

        public static void CreateFromAxisAngle( ref Vector3 axis, float angle, out Matrix result )
        {
            float x = axis.X, y = axis.Y, z = axis.Z;
            float sa = (float)Math.Sin(angle), ca = (float)Math.Cos(angle);
            float xx = x * x, yy = y * y, zz = z * z;
            float xy = x * y, xz = x * z, yz = y * z;

            result.M11 = xx + ca * ( 1.0f - xx );   result.M12 = xy - ca * xy + sa * z;     result.M13 = xz - ca * xz - sa * y;     result.M14 = 0.0f;
            result.M21 = xy - ca * xy - sa * z;     result.M22 = yy + ca * ( 1.0f - yy );   result.M23 = yz - ca * yz + sa * x;     result.M24 = 0.0f;
            result.M31 = xz - ca * xz + sa * y;     result.M32 = yz - ca * yz - sa * x;     result.M33 = zz + ca * ( 1.0f - zz );   result.M34 = 0.0f;
            result.M41 = 0.0f;                      result.M42 = 0.0f;                      result.M43 = 0.0f;                      result.M44 = 1.0f;
        }


        /// <summary>
        /// Builds a perspective projection matrix based on a field of view.
        /// </summary>
        /// <param name="fieldOfView">Field of view in the y direction, in radians.</param>
        /// <param name="aspectRatio">Aspect ratio, defined as view space width divided by height.</param>
        /// <param name="nearPlaneDistance">Distance from the orign to the near view-plane.</param>
        /// <param name="farPlaneDistance">Distance from the origin to the far view-plane.</param>
        /// <returns></returns>
        public static Matrix CreatePerspectiveFieldOfView(float fieldOfView, float aspectRatio, float nearPlaneDistance, float farPlaneDistance)
        {
            if (fieldOfView <= 0.0f || fieldOfView >= MathHelper.Pi)
                throw new ArgumentOutOfRangeException("fieldOfView", String.Format( CultureInfo.CurrentCulture, "FrameworkResources.OutRangeFieldOfView", "fieldOfView" ) );

            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance )
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            float yScale = 1.0f / (float)System.Math.Tan(fieldOfView * 0.5f);
            float xScale = yScale / aspectRatio;

            Matrix result;

            result.M11 = xScale;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = yScale;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M31 = result.M32 = 0.0f;
            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M34 = -1.0f;

            result.M41 = result.M42 = result.M44 = 0.0f;
            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);

            return result;
        }

        public static void CreatePerspectiveFieldOfView(float fieldOfView, float aspectRatio, float nearPlaneDistance, float farPlaneDistance, out Matrix result)
        {
            if (fieldOfView <= 0.0f || fieldOfView >= MathHelper.Pi)
                throw new ArgumentOutOfRangeException("fieldOfView", String.Format( CultureInfo.CurrentCulture, "FrameworkResources.OutRangeFieldOfView", "fieldOfView" ) );

            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance )
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            float yScale = 1.0f / (float)System.Math.Tan(fieldOfView * 0.5f);
            float xScale = yScale / aspectRatio;

            result.M11 = xScale;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = yScale;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M31 = result.M32 = 0.0f;
            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M34 = -1.0f;

            result.M41 = result.M42 = result.M44 = 0.0f;
            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
        }

        /// <summary>
        /// Builds a perspective projection matrix.
        /// </summary>
        /// <param name="width">width of the view volume at the near view-plane.</param>
        /// <param name="height">height of the view volume at the near view-plane.</param>
        /// <param name="nearPlaneDistance">Distance from the orign to the near view-plane.</param>
        /// <param name="farPlaneDistance">Distance from the origin to the far view-plane.</param>
        /// <returns></returns>
        public static Matrix CreatePerspective(float width, float height, float nearPlaneDistance, float farPlaneDistance)
        {
            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            Matrix result;

            result.M11 = 2.0f * nearPlaneDistance / width;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f * nearPlaneDistance / height;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M31 = result.M32 = 0.0f;
            result.M34 = -1.0f;

            result.M41 = result.M42 = result.M44 = 0.0f;
            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);

            return result;
        }

        public static void CreatePerspective(float width, float height, float nearPlaneDistance, float farPlaneDistance, out Matrix result)
        {
            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            result.M11 = 2.0f * nearPlaneDistance / width;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f * nearPlaneDistance / height;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M31 = result.M32 = 0.0f;
            result.M34 = -1.0f;

            result.M41 = result.M42 = result.M44 = 0.0f;
            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
        }

        /// <summary>
        /// Builds a customized, perspective projection matrix.
        /// </summary>
        /// <param name="left">Minimum x-value of the view volume at the near view-plane.</param>
        /// <param name="right">Maximum x-value of the view volume at the near view-plane.</param>
        /// <param name="bottom">Minimum y-value of the view volume at the near view-plane.</param>
        /// <param name="top">Maximum y-value of the view volume at the near view-plane.</param>
        /// <param name="nearPlaneDistance">Distance from the orign to the near view-plane.</param>
        /// <param name="farPlaneDistance">Distance from the origin to the far view-plane.</param>
        /// <returns></returns>
        public static Matrix CreatePerspectiveOffCenter(float left, float right, float bottom, float top, float nearPlaneDistance, float farPlaneDistance)
        {
            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            Matrix result;

            result.M11 = 2.0f * nearPlaneDistance / (right - left);
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f * nearPlaneDistance / (top - bottom);
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M31 = (left + right) / (right - left);
            result.M32 = (top + bottom) / (top - bottom);
            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M34 = -1.0f;

            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M41 = result.M42 = result.M44 = 0.0f;

            return result;
        }

        public static void CreatePerspectiveOffCenter(float left, float right, float bottom, float top, float nearPlaneDistance, float farPlaneDistance, out Matrix result)
        {
            if (nearPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "nearPlaneDistance"));

            if (farPlaneDistance <= 0.0f)
                throw new ArgumentOutOfRangeException("farPlaneDistance", String.Format(CultureInfo.CurrentCulture, "FrameworkResources.NegativePlaneDistance", "farPlaneDistance"));

            if (nearPlaneDistance >= farPlaneDistance)
                throw new ArgumentOutOfRangeException("nearPlaneDistance", "FrameworkResources.OppositePlanes");

            result.M11 = 2.0f * nearPlaneDistance / (right - left);
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f * nearPlaneDistance / (top - bottom);
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M31 = (left + right) / (right - left);
            result.M32 = (top + bottom) / (top - bottom);
            result.M33 = farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M34 = -1.0f;

            result.M43 = nearPlaneDistance * farPlaneDistance / (nearPlaneDistance - farPlaneDistance);
            result.M41 = result.M42 = result.M44 = 0.0f;
        }

        /// <summary>
        /// Builds Orthognal projection matrix.
        /// </summary>
        /// <param name="width">Width of the view volume.</param>
        /// <param name="height">Height of the view volume.</param>
        /// <param name="zNearPlane">Minimum z-value of the view volume.</param>
        /// <param name="zFarPlane">Maximum z-value of the view volume.</param>
        /// <returns></returns>
        public static Matrix CreateOrthographic(float width, float height, float zNearPlane, float zFarPlane)
        {
            Matrix result;

            result.M11 = 2.0f / width;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f / height;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = 1.0f / (zNearPlane - zFarPlane);
            result.M31 = result.M32 = result.M34 = 0.0f;

            result.M41 = result.M42 = 0.0f;
            result.M43 = zNearPlane / (zNearPlane - zFarPlane);
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateOrthographic(float width, float height, float zNearPlane, float zFarPlane, out Matrix result)
        {
            result.M11 = 2.0f / width;
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f / height;
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = 1.0f / (zNearPlane - zFarPlane);
            result.M31 = result.M32 = result.M34 = 0.0f;

            result.M41 = result.M42 = 0.0f;
            result.M43 = zNearPlane / (zNearPlane - zFarPlane);
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Builds a customized, orthognal projection matrix.
        /// </summary>
        /// <param name="left">Minimum x-value of the view volume.</param>
        /// <param name="right">Maximum x-value of the view volume.</param>
        /// <param name="bottom">Minimum y-value of the view volume.</param>
        /// <param name="top">Maximum y-value of the view volume.</param>
        /// <param name="zNearPlane">Minimum z-value of the view volume.</param>
        /// <param name="zFarPlane">Maximum z-value of the view volume.</param>
        /// <returns></returns>
        public static Matrix CreateOrthographicOffCenter(float left, float right, float bottom, float top, float zNearPlane, float zFarPlane)
        {
            Matrix result;

            result.M11 = 2.0f / (right - left);
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f / (top - bottom);
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = 1.0f / (zNearPlane - zFarPlane);
            result.M31 = result.M32 = result.M34 = 0.0f;

            result.M41 = (left + right) / (left - right);
            result.M42 = (top + bottom) / (bottom - top);
            result.M43 = zNearPlane / (zNearPlane - zFarPlane);
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateOrthographicOffCenter(float left, float right, float bottom, float top, float zNearPlane, float zFarPlane, out Matrix result)
        {
            result.M11 = 2.0f / (right - left);
            result.M12 = result.M13 = result.M14 = 0.0f;

            result.M22 = 2.0f / (top - bottom);
            result.M21 = result.M23 = result.M24 = 0.0f;

            result.M33 = 1.0f / (zNearPlane - zFarPlane);
            result.M31 = result.M32 = result.M34 = 0.0f;

            result.M41 = (left + right) / (left - right);
            result.M42 = (top + bottom) / (bottom - top);
            result.M43 = zNearPlane / (zNearPlane - zFarPlane);
            result.M44 = 1.0f;
        }


        /// <summary>
        /// Build look-at matrix.
        /// </summary>
        /// <param name="cameraPosition">Camera position. this value is used in traslation.</param>
        /// <param name="cameraTarget">look-at target position.</param>
        /// <param name="cameraUpVector">current world's up, usually [0, 1, 0].</param>
        /// <returns></returns>
        public static Matrix CreateLookAt(Vector3 cameraPosition, Vector3 cameraTarget, Vector3 cameraUpVector)
        {
            Vector3 zaxis = Vector3.Normalize(cameraPosition - cameraTarget);
            Vector3 xaxis = Vector3.Normalize(Vector3.Cross(cameraUpVector, zaxis));
            Vector3 yaxis = Vector3.Cross(zaxis, xaxis);

            Matrix result;
            result.M11 = xaxis.X; result.M12 = yaxis.X; result.M13 = zaxis.X; result.M14 = 0.0f;
            result.M21 = xaxis.Y; result.M22 = yaxis.Y; result.M23 = zaxis.Y; result.M24 = 0.0f;
            result.M31 = xaxis.Z; result.M32 = yaxis.Z; result.M33 = zaxis.Z; result.M34 = 0.0f;
            result.M41 = -Vector3.Dot(xaxis, cameraPosition);
            result.M42 = -Vector3.Dot(yaxis, cameraPosition);
            result.M43 = -Vector3.Dot(zaxis, cameraPosition);
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateLookAt(ref Vector3 cameraPosition, ref Vector3 cameraTarget, ref Vector3 cameraUpVector, out Matrix result)
        {
            Vector3 zaxis = Vector3.Normalize(cameraPosition - cameraTarget);
            Vector3 xaxis = Vector3.Normalize(Vector3.Cross(cameraUpVector, zaxis));
            Vector3 yaxis = Vector3.Cross(zaxis, xaxis);

            result.M11 = xaxis.X; result.M12 = yaxis.X; result.M13 = zaxis.X; result.M14 = 0.0f;
            result.M21 = xaxis.Y; result.M22 = yaxis.Y; result.M23 = zaxis.Y; result.M24 = 0.0f;
            result.M31 = xaxis.Z; result.M32 = yaxis.Z; result.M33 = zaxis.Z; result.M34 = 0.0f;
            result.M41 = -Vector3.Dot(xaxis, cameraPosition);
            result.M42 = -Vector3.Dot(yaxis, cameraPosition);
            result.M43 = -Vector3.Dot(zaxis, cameraPosition);
            result.M44 = 1.0f;
        }

        /// <summary>
        /// Build world matrix.
        /// </summary>
        /// <param name="position">Object position. this value is used in traslation.</param>
        /// <param name="forward">Forward direction of the object.</param>
        /// <param name="up">current object's up, usually [0, 1, 0].</param>
        /// <returns></returns>
        public static Matrix CreateWorld(Vector3 position, Vector3 forward, Vector3 up)
        {
            Vector3 zaxis = Vector3.Normalize(-forward);
            Vector3 xaxis = Vector3.Normalize(Vector3.Cross(up, zaxis));
            Vector3 yaxis = Vector3.Cross(zaxis, xaxis);

            Matrix result;
            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;
            result.M41 = position.X;
            result.M42 = position.Y;
            result.M43 = position.Z;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateWorld(ref Vector3 position, ref Vector3 forward, ref Vector3 up, out Matrix result)
        {
            Vector3 zaxis = Vector3.Normalize(-forward);
            Vector3 xaxis = Vector3.Normalize(Vector3.Cross(up, zaxis));
            Vector3 yaxis = Vector3.Cross(zaxis, xaxis);

            result.M11 = xaxis.X; result.M12 = xaxis.Y; result.M13 = xaxis.Z; result.M14 = 0.0f;
            result.M21 = yaxis.X; result.M22 = yaxis.Y; result.M23 = yaxis.Z; result.M24 = 0.0f;
            result.M31 = zaxis.X; result.M32 = zaxis.Y; result.M33 = zaxis.Z; result.M34 = 0.0f;
            result.M41 = position.X;
            result.M42 = position.Y;
            result.M43 = position.Z;
            result.M44 = 1.0f;
        }

        /// <summary>
        /// Create Rotation matrix from quaternion.
        /// </summary>
        /// <param name="quaternion"></param>
        /// <returns></returns>
        public static Matrix CreateFromQuaternion(Quaternion quaternion)
        {
            Matrix result;

            float xx = quaternion.X * quaternion.X;
            float yy = quaternion.Y * quaternion.Y;
            float zz = quaternion.Z * quaternion.Z;

            float xy = quaternion.X * quaternion.Y;
            float wz = quaternion.Z * quaternion.W;
            float xz = quaternion.Z * quaternion.X;
            float wy = quaternion.Y * quaternion.W;
            float yz = quaternion.Y * quaternion.Z;
            float wx = quaternion.X * quaternion.W;

            result.M11 = 1.0f - 2.0f * (yy + zz);   result.M12 = 2.0f * (xy + wz);          result.M13 = 2.0f * (xz - wy);          result.M14 = 0.0f;
            result.M21 = 2.0f * (xy - wz);          result.M22 = 1.0f - 2.0f * (zz + xx);   result.M23 = 2.0f * (yz + wx);          result.M24 = 0.0f;
            result.M31 = 2.0f * (xz + wy);          result.M32 = 2.0f * (yz - wx);          result.M33 = 1.0f - 2.0f * (yy + xx);   result.M34 = 0.0f;
            result.M41 = 0.0f;                      result.M42 = 0.0f;                      result.M43 = 0.0f;                      result.M44 = 1.0f;

            return result;
        }

        public static void CreateFromQuaternion(ref Quaternion quaternion, out Matrix result)
        {
            float xx = quaternion.X * quaternion.X;
            float yy = quaternion.Y * quaternion.Y;
            float zz = quaternion.Z * quaternion.Z;

            float xy = quaternion.X * quaternion.Y;
            float wz = quaternion.Z * quaternion.W;
            float xz = quaternion.Z * quaternion.X;
            float wy = quaternion.Y * quaternion.W;
            float yz = quaternion.Y * quaternion.Z;
            float wx = quaternion.X * quaternion.W;

            result.M11 = 1.0f - 2.0f * (yy + zz);   result.M12 = 2.0f * (xy + wz);          result.M13 = 2.0f * (xz - wy);          result.M14 = 0.0f;
            result.M21 = 2.0f * (xy - wz);          result.M22 = 1.0f - 2.0f * (zz + xx);   result.M23 = 2.0f * (yz + wx);          result.M24 = 0.0f;
            result.M31 = 2.0f * (xz + wy);          result.M32 = 2.0f * (yz - wx);          result.M33 = 1.0f - 2.0f * (yy + xx);   result.M34 = 0.0f;
            result.M41 = 0.0f;                      result.M42 = 0.0f;                      result.M43 = 0.0f;                      result.M44 = 1.0f;
        }

        /// <summary>
        /// Create a matrix with a specfied yaw, pitch, and roll.
        /// </summary>
        /// <param name="yaw">Yaw around the y-axis, in radians.</param>
        /// <param name="pitch">Pitch around the x-axis, in radians.</param>
        /// <param name="roll">Roll around the z-axis, in radians.</param>
        /// <returns></returns>
        public static Matrix CreateFromYawPitchRoll( float yaw, float pitch, float roll )
        {
            Quaternion q;
            Quaternion.CreateFromYawPitchRoll(yaw, pitch, roll, out q);

            Matrix m;
            Matrix.CreateFromQuaternion(ref q, out m);
            return m;
        }

        public static void CreateFromYawPitchRoll(float yaw, float pitch, float roll, out Matrix result)
        {
            Quaternion q;
            Quaternion.CreateFromYawPitchRoll(yaw, pitch, roll, out q);
            Matrix.CreateFromQuaternion(ref q, out result);
        }

        /// <summary>
        /// Builds a matrix that flattens geometry into a plane.
        /// </summary>
        /// <param name="lightDirection"></param>
        /// <param name="plane"></param>
        /// <returns></returns>
        /// <remarks>This method flatterns geometry into a plne, as if casting a shadow from a light.</remarks>
        public static Matrix CreateShadow(Vector3 lightDirection, Plane plane)
        {
            Plane p;
            Plane.Normalize(ref plane, out p);

            float dot = p.Normal.X * lightDirection.X + p.Normal.Y * lightDirection.Y + p.Normal.Z * lightDirection.Z;
            float a = -p.Normal.X;
            float b = -p.Normal.Y;
            float c = -p.Normal.Z;
            float d = -p.D;

            Matrix result;
            result.M11 = a * lightDirection.X + dot;
            result.M21 = b * lightDirection.X;
            result.M31 = c * lightDirection.X;
            result.M41 = d * lightDirection.X;

            result.M12 = a * lightDirection.Y;
            result.M22 = b * lightDirection.Y + dot;
            result.M32 = c * lightDirection.Y;
            result.M42 = d * lightDirection.Y;

            result.M13 = a * lightDirection.Z;
            result.M23 = b * lightDirection.Z;
            result.M33 = c * lightDirection.Z + dot;
            result.M43 = d * lightDirection.Z;

            result.M14 = 0.0f;
            result.M24 = 0.0f;
            result.M34 = 0.0f;
            result.M44 = dot;

            return result;
        }

        public static void CreateShadow(ref Vector3 lightDirection, ref Plane plane, out Matrix result)
        {
            Plane p;
            Plane.Normalize(ref plane, out p);

            float dot = p.Normal.X * lightDirection.X + p.Normal.Y * lightDirection.Y + p.Normal.Z * lightDirection.Z;
            float a = -p.Normal.X;
            float b = -p.Normal.Y;
            float c = -p.Normal.Z;
            float d = -p.D;

            result.M11 = a * lightDirection.X + dot;
            result.M21 = b * lightDirection.X;
            result.M31 = c * lightDirection.X;
            result.M41 = d * lightDirection.X;

            result.M12 = a * lightDirection.Y;
            result.M22 = b * lightDirection.Y + dot;
            result.M32 = c * lightDirection.Y;
            result.M42 = d * lightDirection.Y;

            result.M13 = a * lightDirection.Z;
            result.M23 = b * lightDirection.Z;
            result.M33 = c * lightDirection.Z + dot;
            result.M43 = d * lightDirection.Z;

            result.M14 = 0.0f;
            result.M24 = 0.0f;
            result.M34 = 0.0f;
            result.M44 = dot;
        }

        /// <summary>
        /// Create a matrix that reflects the coodinate system about a plane.
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static Matrix CreateReflection(Plane value)
        {
            value.Normalize();

            float a = value.Normal.X;
            float b = value.Normal.Y;
            float c = value.Normal.Z;

            float fa = -2.0f * a;
            float fb = -2.0f * b;
            float fc = -2.0f * c;

            Matrix result;

            result.M11 = fa * a + 1.0f;
            result.M12 = fb * a;
            result.M13 = fc * a;
            result.M14 = 0.0f;

            result.M21 = fa * b;
            result.M22 = fb * b + 1.0f;
            result.M23 = fc * b;
            result.M24 = 0.0f;

            result.M31 = fa * c;
            result.M32 = fb * c;
            result.M33 = fc * c + 1.0f;
            result.M34 = 0.0f;

            result.M41 = fa * value.D;
            result.M42 = fb * value.D;
            result.M43 = fc * value.D;
            result.M44 = 1.0f;

            return result;
        }

        public static void CreateReflection(ref Plane value, out Matrix result)
        {
            Plane p;
            Plane.Normalize(ref value, out p);
            value.Normalize();

            float a = p.Normal.X;
            float b = p.Normal.Y;
            float c = p.Normal.Z;

            float fa = -2.0f * a;
            float fb = -2.0f * b;
            float fc = -2.0f * c;

            result.M11 = fa * a + 1.0f;
            result.M12 = fb * a;
            result.M13 = fc * a;
            result.M14 = 0.0f;

            result.M21 = fa * b;
            result.M22 = fb * b + 1.0f;
            result.M23 = fc * b;
            result.M24 = 0.0f;

            result.M31 = fa * c;
            result.M32 = fb * c;
            result.M33 = fc * c + 1.0f;
            result.M34 = 0.0f;

            result.M41 = fa * p.D;
            result.M42 = fb * p.D;
            result.M43 = fc * p.D;
            result.M44 = 1.0f;
        }

        /// <summary>
        /// Transform a matrix by given quaterion.
        /// </summary>
        /// <param name="value"></param>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Matrix Transform(Matrix value, Quaternion rotation)
        {
            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float q11 = 1.0f - yy2 - zz2;
            float q21 = xy2 - wz2;
            float q31 = xz2 + wy2;

            float q12 = xy2 + wz2;
            float q22 = 1.0f - xx2 - zz2;
            float q32 = yz2 - wx2;

            float q13 = xz2 - wy2;
            float q23 = yz2 + wx2;
            float q33 = 1.0f - xx2 - yy2;

            Matrix result;

            // First row
            result.M11 = value.M11 * q11 + value.M12 * q21 + value.M13 * q31;
            result.M12 = value.M11 * q12 + value.M12 * q22 + value.M13 * q32;
            result.M13 = value.M11 * q13 + value.M12 * q23 + value.M13 * q33;
            result.M14 = value.M14;
            // Second row
            result.M21 = value.M21 * q11 + value.M22 * q21 + value.M23 * q31;
            result.M22 = value.M21 * q12 + value.M22 * q22 + value.M23 * q32;
            result.M23 = value.M21 * q13 + value.M22 * q23 + value.M23 * q33;
            result.M24 = value.M24;
            // Third row
            result.M31 = value.M31 * q11 + value.M32 * q21 + value.M33 * q31;
            result.M32 = value.M31 * q12 + value.M32 * q22 + value.M33 * q32;
            result.M33 = value.M31 * q13 + value.M32 * q23 + value.M33 * q33;
            result.M34 = value.M34;
            // Forth row
            result.M41 = value.M41 * q11 + value.M42 * q21 + value.M43 * q31;
            result.M42 = value.M41 * q12 + value.M42 * q22 + value.M43 * q32;
            result.M43 = value.M41 * q13 + value.M42 * q23 + value.M43 * q33;
            result.M44 = value.M44;

            return result;
        }

        public static void Transform(ref Matrix value, ref Quaternion rotation, out Matrix result)
        {
            // Compute rotation matrix.
            float x2 = rotation.X + rotation.X;
            float y2 = rotation.Y + rotation.Y;
            float z2 = rotation.Z + rotation.Z;

            float wx2 = rotation.W * x2;
            float wy2 = rotation.W * y2;
            float wz2 = rotation.W * z2;
            float xx2 = rotation.X * x2;
            float xy2 = rotation.X * y2;
            float xz2 = rotation.X * z2;
            float yy2 = rotation.Y * y2;
            float yz2 = rotation.Y * z2;
            float zz2 = rotation.Z * z2;

            float q11 = 1.0f - yy2 - zz2;
            float q21 = xy2 - wz2;
            float q31 = xz2 + wy2;

            float q12 = xy2 + wz2;
            float q22 = 1.0f - xx2 - zz2;
            float q32 = yz2 - wx2;

            float q13 = xz2 - wy2;
            float q23 = yz2 + wx2;
            float q33 = 1.0f - xx2 - yy2;

            // First row
            float m11 = value.M11 * q11 + value.M12 * q21 + value.M13 * q31;
            float m12 = value.M11 * q12 + value.M12 * q22 + value.M13 * q32;
            float m13 = value.M11 * q13 + value.M12 * q23 + value.M13 * q33;
            float m14 = value.M14;
            // Second row
            float m21 = value.M21 * q11 + value.M22 * q21 + value.M23 * q31;
            float m22 = value.M21 * q12 + value.M22 * q22 + value.M23 * q32;
            float m23 = value.M21 * q13 + value.M22 * q23 + value.M23 * q33;
            float m24 = value.M24;
            // Third row
            float m31 = value.M31 * q11 + value.M32 * q21 + value.M33 * q31;
            float m32 = value.M31 * q12 + value.M32 * q22 + value.M33 * q32;
            float m33 = value.M31 * q13 + value.M32 * q23 + value.M33 * q33;
            float m34 = value.M34;
            // Forth row
            float m41 = value.M41 * q11 + value.M42 * q21 + value.M43 * q31;
            float m42 = value.M41 * q12 + value.M42 * q22 + value.M43 * q32;
            float m43 = value.M41 * q13 + value.M42 * q23 + value.M43 * q33;
            float m44 = value.M44;

            // copy back to result.
            result.M11 = m11; result.M12 = m12; result.M13 = m13; result.M14 = m14;
            result.M21 = m21; result.M22 = m22; result.M23 = m23; result.M24 = m24;
            result.M31 = m31; result.M32 = m32; result.M33 = m33; result.M34 = m34;
            result.M41 = m41; result.M42 = m42; result.M43 = m43; result.M44 = m44; 
        }
        #endregion

        #region Overrides methods
        public override string ToString()
        {
            CultureInfo ci = CultureInfo.CurrentCulture;
            return
                "{ " +
                String.Format(ci, "{{M11:{0} M12:{1} M13:{2} M14:{3}}} ", M11.ToString(ci), M12.ToString(ci), M13.ToString(ci), M14.ToString(ci)) +
                String.Format(ci, "{{M21:{0} M22:{1} M23:{2} M24:{3}}} ", M21.ToString(ci), M22.ToString(ci), M23.ToString(ci), M24.ToString(ci)) +
                String.Format(ci, "{{M31:{0} M32:{1} M33:{2} M34:{3}}} ", M31.ToString(ci), M32.ToString(ci), M33.ToString(ci), M34.ToString(ci)) +
                String.Format(ci, "{{M41:{0} M42:{1} M43:{2} M44:{3}}} ", M41.ToString(ci), M42.ToString(ci), M43.ToString(ci), M44.ToString(ci)) +
                "}";
        }

        // Implement IEquatable<T>
        public bool Equals(Matrix other)
        {
            return (
                M11 == other.M11 && M22 == other.M22 && M33 == other.M33 && M44 == other.M44 && // Check Diagnal element first for early out.
                                    M12 == other.M12 && M13 == other.M13 && M14 == other.M14 &&
                M21 == other.M21                     && M23 == other.M23 && M24 == other.M24 &&
                M31 == other.M31 && M32 == other.M32                     && M34 == other.M34 &&
                M41 == other.M41 && M42 == other.M42 && M43 == other.M43                        );
        }

        public override bool Equals(object obj)
        {
            bool equal = false;
            if (obj is Matrix)
            {
                equal = Equals((Matrix)obj);
            }
            return equal;
        }

        public override int GetHashCode()
        {
            return M11.GetHashCode() + M12.GetHashCode() + M13.GetHashCode() + M14.GetHashCode() +
                    M21.GetHashCode() + M22.GetHashCode() + M23.GetHashCode() + M24.GetHashCode() +
                    M31.GetHashCode() + M32.GetHashCode() + M33.GetHashCode() + M34.GetHashCode() +
                    M41.GetHashCode() + M42.GetHashCode() + M43.GetHashCode() + M44.GetHashCode();
        }
        #endregion

        #region Operations
        // Unary Operators.
        public static Matrix Transpose(Matrix matrix)
        {
            Matrix result;

            result.M11 = matrix.M11; result.M12 = matrix.M21; result.M13 = matrix.M31; result.M14 = matrix.M41;
            result.M21 = matrix.M12; result.M22 = matrix.M22; result.M23 = matrix.M32; result.M24 = matrix.M42;
            result.M31 = matrix.M13; result.M32 = matrix.M23; result.M33 = matrix.M33; result.M34 = matrix.M43;
            result.M41 = matrix.M14; result.M42 = matrix.M24; result.M43 = matrix.M34; result.M44 = matrix.M44;

            return result;
        }

        public static void Transpose(ref Matrix matrix, out Matrix result)
        {
            float m11, m12, m13, m14;
            float m21, m22, m23, m24;
            float m31, m32, m33, m34;
            float m41, m42, m43, m44;

            m11 = matrix.M11; m12 = matrix.M12; m13 = matrix.M13; m14 = matrix.M14;
            m21 = matrix.M21; m22 = matrix.M22; m23 = matrix.M23; m24 = matrix.M24;
            m31 = matrix.M31; m32 = matrix.M32; m33 = matrix.M33; m34 = matrix.M34;
            m41 = matrix.M41; m42 = matrix.M42; m43 = matrix.M43; m44 = matrix.M44;

            result.M11 = m11; result.M12 = m21; result.M13 = m31; result.M14 = m41;
            result.M21 = m12; result.M22 = m22; result.M23 = m32; result.M24 = m42;
            result.M31 = m13; result.M32 = m23; result.M33 = m33; result.M34 = m43;
            result.M41 = m14; result.M42 = m24; result.M43 = m34; result.M44 = m44;
        }


        public float Determinant()
        {
            // | a b c d |     | f g h |     | e g h |     | e f h |     | e f g |
            // | e f g h | = a | j k l | - b | i k l | + c | i j l | - d | i j k |
            // | i j k l |     | n o p |     | m o p |     | m n p |     | m n o |
            // | m n o p |
            //
            //   | f g h |
            // a | j k l | = a ( f ( kp - lo ) - g ( jp - ln ) + h ( jo - kn ) )
            //   | n o p |
            //
            //   | e g h |     
            // b | i k l | = b ( e ( kp - lo ) - g ( ip - lm ) + h ( io - km ) )
            //   | m o p |     
            //
            //   | e f h |
            // c | i j l | = c ( e ( jp - ln ) - f ( ip - lm ) + h ( in - jm ) )
            //   | m n p |
            //
            //   | e f g |
            // d | i j k | = d ( e ( jo - kn ) - f ( io - km ) + g ( in - jm ) )
            //   | m n o |
            //
            // Cost of operation
            // 17 adds and 28 muls.
            //
            // add: 6 + 8 + 3 = 17
            // mul: 12 + 16 = 28

            float a = M11, b = M12, c = M13, d = M14;
            float e = M21, f = M22, g = M23, h = M24;
            float i = M31, j = M32, k = M33, l = M34;
            float m = M41, n = M42, o = M43, p = M44;

            float kp_lo = k * p - l * o;
            float jp_ln = j * p - l * n;
            float jo_kn = j * o - k * n;
            float ip_lm = i * p - l * m;
            float io_km = i * o - k * m;
            float in_jm = i * n - j * m;

            return a * (f * kp_lo - g * jp_ln + h * jo_kn) -
                   b * (e * kp_lo - g * ip_lm + h * io_km) +
                   c * (e * jp_ln - f * ip_lm + h * in_jm) -
                   d * (e * jo_kn - f * io_km + g * in_jm);
        }

        public static Matrix Invert(Matrix matrix)
        {
            //                                       -1
            // If you have matrix M, inverse Matrix M   can compute
            //
            //     -1       1      
            //    M   = --------- A
            //            det(M)
            //
            // A is adjugate (adjoint) of M, where,
            //
            //      T
            // A = C
            //
            // C is Cofactor matrix of M, where,
            //           i + j
            // C   = (-1)      * det(M  )
            //  ij                    ij
            //
            //     [ a b c d ]
            // M = [ e f g h ]
            //     [ i j k l ]
            //     [ m n o p ]
            //
            // First Row
            //           2 | f g h |
            // C   = (-1)  | j k l | = + ( f ( kp - lo ) - g ( jp - ln ) + h ( jo - kn ) )
            //  11         | n o p |
            //
            //           3 | e g h |
            // C   = (-1)  | i k l | = - ( e ( kp - lo ) - g ( ip - lm ) + h ( io - km ) )
            //  12         | m o p |
            //
            //           4 | e f h |
            // C   = (-1)  | i j l | = + ( e ( jp - ln ) - f ( ip - lm ) + h ( in - jm ) )
            //  13         | m n p |
            //
            //           5 | e f g |
            // C   = (-1)  | i j k | = - ( e ( jo - kn ) - f ( io - km ) + g ( in - jm ) )
            //  14         | m n o |
            //
            // Second Row
            //           3 | b c d |
            // C   = (-1)  | j k l | = - ( b ( kp - lo ) - c ( jp - ln ) + d ( jo - kn ) )
            //  21         | n o p |
            //
            //           4 | a c d |
            // C   = (-1)  | i k l | = + ( a ( kp - lo ) - c ( ip - lm ) + d ( io - km ) )
            //  22         | m o p |
            //
            //           5 | a b d |
            // C   = (-1)  | i j l | = - ( a ( jp - ln ) - b ( ip - lm ) + d ( in - jm ) )
            //  23         | m n p |
            //
            //           6 | a b c |
            // C   = (-1)  | i j k | = + ( a ( jo - kn ) - b ( io - km ) + c ( in - jm ) )
            //  24         | m n o |
            //
            // Third Row
            //           4 | b c d |
            // C   = (-1)  | f g h | = + ( b ( gp - ho ) - c ( fp - hn ) + d ( fo - gn ) )
            //  31         | n o p |
            //
            //           5 | a c d |
            // C   = (-1)  | e g h | = - ( a ( gp - ho ) - c ( ep - hm ) + d ( eo - gm ) )
            //  32         | m o p |
            //
            //           6 | a b d |
            // C   = (-1)  | e f h | = + ( a ( fp - hn ) - b ( ep - hm ) + d ( en - fm ) )
            //  33         | m n p |
            //
            //           7 | a b c |
            // C   = (-1)  | e f g | = - ( a ( fo - gn ) - b ( eo - gm ) + c ( en - fm ) )
            //  34         | m n o |
            //
            // Forth Row
            //           5 | b c d |
            // C   = (-1)  | f g h | = - ( b ( gl - hk ) - c ( fl - hj ) + d ( fk - gj ) )
            //  41         | j k l |
            //
            //           6 | a c d |
            // C   = (-1)  | e g h | = + ( a ( gl - hk ) - c ( el - hi ) + d ( ek - gi ) )
            //  42         | i k l |
            //
            //           7 | a b d |
            // C   = (-1)  | e f h | = - ( a ( fl - hj ) - b ( el - hi ) + d ( ej - fi ) )
            //  43         | i j l |
            //
            //           8 | a b c |
            // C   = (-1)  | e f g | = + ( a ( fk - gj ) - b ( ek - gi ) + c ( ej - fi ) )
            //  44         | i j k |
            //
            // Cost of operation
            // 53 adds, 88 muls, and 1 div.
            //
            // add: 18+32+3 = 53
            // mul: 36+64+4 = 104
            // div: 1
            float a = matrix.M11, b = matrix.M12, c = matrix.M13, d = matrix.M14;
            float e = matrix.M21, f = matrix.M22, g = matrix.M23, h = matrix.M24;
            float i = matrix.M31, j = matrix.M32, k = matrix.M33, l = matrix.M34;
            float m = matrix.M41, n = matrix.M42, o = matrix.M43, p = matrix.M44;

            float kp_lo = k * p - l * o;
            float jp_ln = j * p - l * n;
            float jo_kn = j * o - k * n;
            float ip_lm = i * p - l * m;
            float io_km = i * o - k * m;
            float in_jm = i * n - j * m;

            Matrix result;

            float a11 = +(f * kp_lo - g * jp_ln + h * jo_kn);
            float a12 = -(e * kp_lo - g * ip_lm + h * io_km);
            float a13 = +(e * jp_ln - f * ip_lm + h * in_jm);
            float a14 = -(e * jo_kn - f * io_km + g * in_jm);

            float invDet = 1.0f / (a * a11 + b * a12 + c * a13 + d * a14);

            result.M11 = a11 * invDet;
            result.M21 = a12 * invDet;
            result.M31 = a13 * invDet;
            result.M41 = a14 * invDet;

            result.M12 = -(b * kp_lo - c * jp_ln + d * jo_kn) * invDet;
            result.M22 = +(a * kp_lo - c * ip_lm + d * io_km) * invDet;
            result.M32 = -(a * jp_ln - b * ip_lm + d * in_jm) * invDet;
            result.M42 = +(a * jo_kn - b * io_km + c * in_jm) * invDet;

            float gp_ho = g * p - h * o;
            float fp_hn = f * p - h * n;
            float fo_gn = f * o - g * n;
            float ep_hm = e * p - h * m;
            float eo_gm = e * o - g * m;
            float en_fm = e * n - f * m;

            result.M13 = +(b * gp_ho - c * fp_hn + d * fo_gn) * invDet;
            result.M23 = -(a * gp_ho - c * ep_hm + d * eo_gm) * invDet;
            result.M33 = +(a * fp_hn - b * ep_hm + d * en_fm) * invDet;
            result.M43 = -(a * fo_gn - b * eo_gm + c * en_fm) * invDet;

            float gl_hk = g * l - h * k;
            float fl_hj = f * l - h * j;
            float fk_gj = f * k - g * j;
            float el_hi = e * l - h * i;
            float ek_gi = e * k - g * i;
            float ej_fi = e * j - f * i;

            result.M14 = -(b * gl_hk - c * fl_hj + d * fk_gj) * invDet;
            result.M24 = +(a * gl_hk - c * el_hi + d * ek_gi) * invDet;
            result.M34 = -(a * fl_hj - b * el_hi + d * ej_fi) * invDet;
            result.M44 = +(a * fk_gj - b * ek_gi + c * ej_fi) * invDet;

            return result;
        }

        public static void Invert(ref Matrix matrix, out Matrix result)
        {
            float a = matrix.M11, b = matrix.M12, c = matrix.M13, d = matrix.M14;
            float e = matrix.M21, f = matrix.M22, g = matrix.M23, h = matrix.M24;
            float i = matrix.M31, j = matrix.M32, k = matrix.M33, l = matrix.M34;
            float m = matrix.M41, n = matrix.M42, o = matrix.M43, p = matrix.M44;

            float kp_lo = k * p - l * o;
            float jp_ln = j * p - l * n;
            float jo_kn = j * o - k * n;
            float ip_lm = i * p - l * m;
            float io_km = i * o - k * m;
            float in_jm = i * n - j * m;

            float a11 = +(f * kp_lo - g * jp_ln + h * jo_kn);
            float a12 = -(e * kp_lo - g * ip_lm + h * io_km);
            float a13 = +(e * jp_ln - f * ip_lm + h * in_jm);
            float a14 = -(e * jo_kn - f * io_km + g * in_jm);

            float invDet = 1.0f / (a * a11 + b * a12 + c * a13 + d * a14);

            result.M11 = a11 * invDet;
            result.M21 = a12 * invDet;
            result.M31 = a13 * invDet;
            result.M41 = a14 * invDet;

            result.M12 = -(b * kp_lo - c * jp_ln + d * jo_kn) * invDet;
            result.M22 = +(a * kp_lo - c * ip_lm + d * io_km) * invDet;
            result.M32 = -(a * jp_ln - b * ip_lm + d * in_jm) * invDet;
            result.M42 = +(a * jo_kn - b * io_km + c * in_jm) * invDet;

            float gp_ho = g * p - h * o;
            float fp_hn = f * p - h * n;
            float fo_gn = f * o - g * n;
            float ep_hm = e * p - h * m;
            float eo_gm = e * o - g * m;
            float en_fm = e * n - f * m;

            result.M13 = +(b * gp_ho - c * fp_hn + d * fo_gn) * invDet;
            result.M23 = -(a * gp_ho - c * ep_hm + d * eo_gm) * invDet;
            result.M33 = +(a * fp_hn - b * ep_hm + d * en_fm) * invDet;
            result.M43 = -(a * fo_gn - b * eo_gm + c * en_fm) * invDet;

            float gl_hk = g * l - h * k;
            float fl_hj = f * l - h * j;
            float fk_gj = f * k - g * j;
            float el_hi = e * l - h * i;
            float ek_gi = e * k - g * i;
            float ej_fi = e * j - f * i;

            result.M14 = -(b * gl_hk - c * fl_hj + d * fk_gj) * invDet;
            result.M24 = +(a * gl_hk - c * el_hi + d * ek_gi) * invDet;
            result.M34 = -(a * fl_hj - b * el_hi + d * ej_fi) * invDet;
            result.M44 = +(a * fk_gj - b * ek_gi + c * ej_fi) * invDet;
        }


        // Binary Operators
        public static Matrix Lerp(Matrix matrix1, Matrix matrix2, float amount)
        {
            Matrix result;
            // First Row
            result.M11 = matrix1.M11 + (matrix2.M11 - matrix1.M11) * amount;
            result.M12 = matrix1.M12 + (matrix2.M12 - matrix1.M12) * amount;
            result.M13 = matrix1.M13 + (matrix2.M13 - matrix1.M13) * amount;
            result.M14 = matrix1.M14 + (matrix2.M14 - matrix1.M14) * amount;
            // Second Row
            result.M21 = matrix1.M21 + (matrix2.M21 - matrix1.M21) * amount;
            result.M22 = matrix1.M22 + (matrix2.M22 - matrix1.M22) * amount;
            result.M23 = matrix1.M23 + (matrix2.M23 - matrix1.M23) * amount;
            result.M24 = matrix1.M24 + (matrix2.M24 - matrix1.M24) * amount;
            // Third Row
            result.M31 = matrix1.M31 + (matrix2.M31 - matrix1.M31) * amount;
            result.M32 = matrix1.M32 + (matrix2.M32 - matrix1.M32) * amount;
            result.M33 = matrix1.M33 + (matrix2.M33 - matrix1.M33) * amount;
            result.M34 = matrix1.M34 + (matrix2.M34 - matrix1.M34) * amount;
            // Forth Row
            result.M41 = matrix1.M41 + (matrix2.M41 - matrix1.M41) * amount;
            result.M42 = matrix1.M42 + (matrix2.M42 - matrix1.M42) * amount;
            result.M43 = matrix1.M43 + (matrix2.M43 - matrix1.M43) * amount;
            result.M44 = matrix1.M44 + (matrix2.M44 - matrix1.M44) * amount;

            return result;
        }

        public static void Lerp(ref Matrix matrix1, ref Matrix matrix2, float amount, out Matrix result)
        {
            // First Row
            result.M11 = matrix1.M11 + (matrix2.M11 - matrix1.M11) * amount;
            result.M12 = matrix1.M12 + (matrix2.M12 - matrix1.M12) * amount;
            result.M13 = matrix1.M13 + (matrix2.M13 - matrix1.M13) * amount;
            result.M14 = matrix1.M14 + (matrix2.M14 - matrix1.M14) * amount;
            // Second Row
            result.M21 = matrix1.M21 + (matrix2.M21 - matrix1.M21) * amount;
            result.M22 = matrix1.M22 + (matrix2.M22 - matrix1.M22) * amount;
            result.M23 = matrix1.M23 + (matrix2.M23 - matrix1.M23) * amount;
            result.M24 = matrix1.M24 + (matrix2.M24 - matrix1.M24) * amount;
            // Third Row
            result.M31 = matrix1.M31 + (matrix2.M31 - matrix1.M31) * amount;
            result.M32 = matrix1.M32 + (matrix2.M32 - matrix1.M32) * amount;
            result.M33 = matrix1.M33 + (matrix2.M33 - matrix1.M33) * amount;
            result.M34 = matrix1.M34 + (matrix2.M34 - matrix1.M34) * amount;
            // Forth Row
            result.M41 = matrix1.M41 + (matrix2.M41 - matrix1.M41) * amount;
            result.M42 = matrix1.M42 + (matrix2.M42 - matrix1.M42) * amount;
            result.M43 = matrix1.M43 + (matrix2.M43 - matrix1.M43) * amount;
            result.M44 = matrix1.M44 + (matrix2.M44 - matrix1.M44) * amount;
        }

        #endregion

        #region Operator overloads equivalent methods
        // Unary
        /// <summary>
        /// Negate individual elements of given matrix
        /// </summary>
        /// <param name="matrix"></param>
        /// <returns></returns>
        public static Matrix Negate(Matrix matrix)
        {
            Matrix result;

            result.M11 = -matrix.M11; result.M12 = -matrix.M12; result.M13 = -matrix.M13; result.M14 = -matrix.M14;
            result.M21 = -matrix.M21; result.M22 = -matrix.M22; result.M23 = -matrix.M23; result.M24 = -matrix.M24;
            result.M31 = -matrix.M31; result.M32 = -matrix.M32; result.M33 = -matrix.M33; result.M34 = -matrix.M34;
            result.M41 = -matrix.M41; result.M42 = -matrix.M42; result.M43 = -matrix.M43; result.M44 = -matrix.M44;

            return result;
        }

        public static void Negate(ref Matrix matrix, out Matrix result)
        {
            result.M11 = -matrix.M11; result.M12 = -matrix.M12; result.M13 = -matrix.M13; result.M14 = -matrix.M14;
            result.M21 = -matrix.M21; result.M22 = -matrix.M22; result.M23 = -matrix.M23; result.M24 = -matrix.M24;
            result.M31 = -matrix.M31; result.M32 = -matrix.M32; result.M33 = -matrix.M33; result.M34 = -matrix.M34;
            result.M41 = -matrix.M41; result.M42 = -matrix.M42; result.M43 = -matrix.M43; result.M44 = -matrix.M44;
        }


        // Binary
        /// <summary>
        /// Add individual elements of given matrix a and b.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        /// <returns></returns>
        public static Matrix Add(Matrix matrix1, Matrix matrix2)
        {
            Matrix result;

            result.M11 = matrix1.M11 + matrix2.M11; result.M12 = matrix1.M12 + matrix2.M12; result.M13 = matrix1.M13 + matrix2.M13; result.M14 = matrix1.M14 + matrix2.M14;
            result.M21 = matrix1.M21 + matrix2.M21; result.M22 = matrix1.M22 + matrix2.M22; result.M23 = matrix1.M23 + matrix2.M23; result.M24 = matrix1.M24 + matrix2.M24;
            result.M31 = matrix1.M31 + matrix2.M31; result.M32 = matrix1.M32 + matrix2.M32; result.M33 = matrix1.M33 + matrix2.M33; result.M34 = matrix1.M34 + matrix2.M34;
            result.M41 = matrix1.M41 + matrix2.M41; result.M42 = matrix1.M42 + matrix2.M42; result.M43 = matrix1.M43 + matrix2.M43; result.M44 = matrix1.M44 + matrix2.M44;

            return result;
        }

        public static void Add(ref Matrix matrix1, ref Matrix matrix2, out Matrix result)
        {
            result.M11 = matrix1.M11 + matrix2.M11; result.M12 = matrix1.M12 + matrix2.M12; result.M13 = matrix1.M13 + matrix2.M13; result.M14 = matrix1.M14 + matrix2.M14;
            result.M21 = matrix1.M21 + matrix2.M21; result.M22 = matrix1.M22 + matrix2.M22; result.M23 = matrix1.M23 + matrix2.M23; result.M24 = matrix1.M24 + matrix2.M24;
            result.M31 = matrix1.M31 + matrix2.M31; result.M32 = matrix1.M32 + matrix2.M32; result.M33 = matrix1.M33 + matrix2.M33; result.M34 = matrix1.M34 + matrix2.M34;
            result.M41 = matrix1.M41 + matrix2.M41; result.M42 = matrix1.M42 + matrix2.M42; result.M43 = matrix1.M43 + matrix2.M43; result.M44 = matrix1.M44 + matrix2.M44;
        }


        /// <summary>
        /// Subtract individual elements of given matrix matrix2 from matrix1.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        /// <returns></returns>
        public static Matrix Subtract(Matrix matrix1, Matrix matrix2)
        {
            Matrix result;

            result.M11 = matrix1.M11 - matrix2.M11; result.M12 = matrix1.M12 - matrix2.M12; result.M13 = matrix1.M13 - matrix2.M13; result.M14 = matrix1.M14 - matrix2.M14;
            result.M21 = matrix1.M21 - matrix2.M21; result.M22 = matrix1.M22 - matrix2.M22; result.M23 = matrix1.M23 - matrix2.M23; result.M24 = matrix1.M24 - matrix2.M24;
            result.M31 = matrix1.M31 - matrix2.M31; result.M32 = matrix1.M32 - matrix2.M32; result.M33 = matrix1.M33 - matrix2.M33; result.M34 = matrix1.M34 - matrix2.M34;
            result.M41 = matrix1.M41 - matrix2.M41; result.M42 = matrix1.M42 - matrix2.M42; result.M43 = matrix1.M43 - matrix2.M43; result.M44 = matrix1.M44 - matrix2.M44;

            return result;
        }

        public static void Subtract(ref Matrix matrix1, ref Matrix matrix2, out Matrix result)
        {
            result.M11 = matrix1.M11 - matrix2.M11; result.M12 = matrix1.M12 - matrix2.M12; result.M13 = matrix1.M13 - matrix2.M13; result.M14 = matrix1.M14 - matrix2.M14;
            result.M21 = matrix1.M21 - matrix2.M21; result.M22 = matrix1.M22 - matrix2.M22; result.M23 = matrix1.M23 - matrix2.M23; result.M24 = matrix1.M24 - matrix2.M24;
            result.M31 = matrix1.M31 - matrix2.M31; result.M32 = matrix1.M32 - matrix2.M32; result.M33 = matrix1.M33 - matrix2.M33; result.M34 = matrix1.M34 - matrix2.M34;
            result.M41 = matrix1.M41 - matrix2.M41; result.M42 = matrix1.M42 - matrix2.M42; result.M43 = matrix1.M43 - matrix2.M43; result.M44 = matrix1.M44 - matrix2.M44;
        }


        /// <summary>
        /// Multiply given matrix matrix1 and matrix2.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        /// <returns></returns>
        public static Matrix Multiply(Matrix matrix1, Matrix matrix2)
        {
            Matrix result;

            // First row
            result.M11 = matrix1.M11 * matrix2.M11 + matrix1.M12 * matrix2.M21 + matrix1.M13 * matrix2.M31 + matrix1.M14 * matrix2.M41;
            result.M12 = matrix1.M11 * matrix2.M12 + matrix1.M12 * matrix2.M22 + matrix1.M13 * matrix2.M32 + matrix1.M14 * matrix2.M42;
            result.M13 = matrix1.M11 * matrix2.M13 + matrix1.M12 * matrix2.M23 + matrix1.M13 * matrix2.M33 + matrix1.M14 * matrix2.M43;
            result.M14 = matrix1.M11 * matrix2.M14 + matrix1.M12 * matrix2.M24 + matrix1.M13 * matrix2.M34 + matrix1.M14 * matrix2.M44;
            // Second row
            result.M21 = matrix1.M21 * matrix2.M11 + matrix1.M22 * matrix2.M21 + matrix1.M23 * matrix2.M31 + matrix1.M24 * matrix2.M41;
            result.M22 = matrix1.M21 * matrix2.M12 + matrix1.M22 * matrix2.M22 + matrix1.M23 * matrix2.M32 + matrix1.M24 * matrix2.M42;
            result.M23 = matrix1.M21 * matrix2.M13 + matrix1.M22 * matrix2.M23 + matrix1.M23 * matrix2.M33 + matrix1.M24 * matrix2.M43;
            result.M24 = matrix1.M21 * matrix2.M14 + matrix1.M22 * matrix2.M24 + matrix1.M23 * matrix2.M34 + matrix1.M24 * matrix2.M44;
            // Third row
            result.M31 = matrix1.M31 * matrix2.M11 + matrix1.M32 * matrix2.M21 + matrix1.M33 * matrix2.M31 + matrix1.M34 * matrix2.M41;
            result.M32 = matrix1.M31 * matrix2.M12 + matrix1.M32 * matrix2.M22 + matrix1.M33 * matrix2.M32 + matrix1.M34 * matrix2.M42;
            result.M33 = matrix1.M31 * matrix2.M13 + matrix1.M32 * matrix2.M23 + matrix1.M33 * matrix2.M33 + matrix1.M34 * matrix2.M43;
            result.M34 = matrix1.M31 * matrix2.M14 + matrix1.M32 * matrix2.M24 + matrix1.M33 * matrix2.M34 + matrix1.M34 * matrix2.M44;
            // Forth row
            result.M41 = matrix1.M41 * matrix2.M11 + matrix1.M42 * matrix2.M21 + matrix1.M43 * matrix2.M31 + matrix1.M44 * matrix2.M41;
            result.M42 = matrix1.M41 * matrix2.M12 + matrix1.M42 * matrix2.M22 + matrix1.M43 * matrix2.M32 + matrix1.M44 * matrix2.M42;
            result.M43 = matrix1.M41 * matrix2.M13 + matrix1.M42 * matrix2.M23 + matrix1.M43 * matrix2.M33 + matrix1.M44 * matrix2.M43;
            result.M44 = matrix1.M41 * matrix2.M14 + matrix1.M42 * matrix2.M24 + matrix1.M43 * matrix2.M34 + matrix1.M44 * matrix2.M44;

            return result;
        }

        public static void Multiply(ref Matrix matrix1, ref Matrix matrix2, out Matrix result)
        {
            // First row
            float m11 = matrix1.M11 * matrix2.M11 + matrix1.M12 * matrix2.M21 + matrix1.M13 * matrix2.M31 + matrix1.M14 * matrix2.M41;
            float m12 = matrix1.M11 * matrix2.M12 + matrix1.M12 * matrix2.M22 + matrix1.M13 * matrix2.M32 + matrix1.M14 * matrix2.M42;
            float m13 = matrix1.M11 * matrix2.M13 + matrix1.M12 * matrix2.M23 + matrix1.M13 * matrix2.M33 + matrix1.M14 * matrix2.M43;
            float m14 = matrix1.M11 * matrix2.M14 + matrix1.M12 * matrix2.M24 + matrix1.M13 * matrix2.M34 + matrix1.M14 * matrix2.M44;
            // Second row
            float m21 = matrix1.M21 * matrix2.M11 + matrix1.M22 * matrix2.M21 + matrix1.M23 * matrix2.M31 + matrix1.M24 * matrix2.M41;
            float m22 = matrix1.M21 * matrix2.M12 + matrix1.M22 * matrix2.M22 + matrix1.M23 * matrix2.M32 + matrix1.M24 * matrix2.M42;
            float m23 = matrix1.M21 * matrix2.M13 + matrix1.M22 * matrix2.M23 + matrix1.M23 * matrix2.M33 + matrix1.M24 * matrix2.M43;
            float m24 = matrix1.M21 * matrix2.M14 + matrix1.M22 * matrix2.M24 + matrix1.M23 * matrix2.M34 + matrix1.M24 * matrix2.M44;
            // Third row
            float m31 = matrix1.M31 * matrix2.M11 + matrix1.M32 * matrix2.M21 + matrix1.M33 * matrix2.M31 + matrix1.M34 * matrix2.M41;
            float m32 = matrix1.M31 * matrix2.M12 + matrix1.M32 * matrix2.M22 + matrix1.M33 * matrix2.M32 + matrix1.M34 * matrix2.M42;
            float m33 = matrix1.M31 * matrix2.M13 + matrix1.M32 * matrix2.M23 + matrix1.M33 * matrix2.M33 + matrix1.M34 * matrix2.M43;
            float m34 = matrix1.M31 * matrix2.M14 + matrix1.M32 * matrix2.M24 + matrix1.M33 * matrix2.M34 + matrix1.M34 * matrix2.M44;
            // Forth row
            float m41 = matrix1.M41 * matrix2.M11 + matrix1.M42 * matrix2.M21 + matrix1.M43 * matrix2.M31 + matrix1.M44 * matrix2.M41;
            float m42 = matrix1.M41 * matrix2.M12 + matrix1.M42 * matrix2.M22 + matrix1.M43 * matrix2.M32 + matrix1.M44 * matrix2.M42;
            float m43 = matrix1.M41 * matrix2.M13 + matrix1.M42 * matrix2.M23 + matrix1.M43 * matrix2.M33 + matrix1.M44 * matrix2.M43;
            float m44 = matrix1.M41 * matrix2.M14 + matrix1.M42 * matrix2.M24 + matrix1.M43 * matrix2.M34 + matrix1.M44 * matrix2.M44;

            // copy back to result.
            result.M11 = m11; result.M12 = m12; result.M13 = m13; result.M14 = m14; 
            result.M21 = m21; result.M22 = m22; result.M23 = m23; result.M24 = m24; 
            result.M31 = m31; result.M32 = m32; result.M33 = m33; result.M34 = m34; 
            result.M41 = m41; result.M42 = m42; result.M43 = m43; result.M44 = m44; 
        }


        /// <summary>
        /// Multiply given factor aginst given matrix.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="scaleFactor"></param>
        /// <returns></returns>
        public static Matrix Multiply(Matrix matrix1, float scaleFactor)
        {
            Matrix result;

            float s = scaleFactor;
            result.M11 = matrix1.M11 * s; result.M12 = matrix1.M12 * s; result.M13 = matrix1.M13 * s; result.M14 = matrix1.M14 * s;
            result.M21 = matrix1.M21 * s; result.M22 = matrix1.M22 * s; result.M23 = matrix1.M23 * s; result.M24 = matrix1.M24 * s;
            result.M31 = matrix1.M31 * s; result.M32 = matrix1.M32 * s; result.M33 = matrix1.M33 * s; result.M34 = matrix1.M34 * s;
            result.M41 = matrix1.M41 * s; result.M42 = matrix1.M42 * s; result.M43 = matrix1.M43 * s; result.M44 = matrix1.M44 * s;

            return result;
        }

        public static void Multiply(ref Matrix matrix1, float scaleFactor, out Matrix result)
        {
            float s = scaleFactor;
            result.M11 = matrix1.M11 * s; result.M12 = matrix1.M12 * s; result.M13 = matrix1.M13 * s; result.M14 = matrix1.M14 * s;
            result.M21 = matrix1.M21 * s; result.M22 = matrix1.M22 * s; result.M23 = matrix1.M23 * s; result.M24 = matrix1.M24 * s;
            result.M31 = matrix1.M31 * s; result.M32 = matrix1.M32 * s; result.M33 = matrix1.M33 * s; result.M34 = matrix1.M34 * s;
            result.M41 = matrix1.M41 * s; result.M42 = matrix1.M42 * s; result.M43 = matrix1.M43 * s; result.M44 = matrix1.M44 * s;
        }


        /// <summary>
        /// Divide individual elements of given matrix matrix1 by matrix matrix2.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        /// <returns></returns>
        public static Matrix Divide(Matrix matrix1, Matrix matrix2)
        {
            Matrix result;

            result.M11 = matrix1.M11 / matrix2.M11; result.M12 = matrix1.M12 / matrix2.M12; result.M13 = matrix1.M13 / matrix2.M13; result.M14 = matrix1.M14 / matrix2.M14;
            result.M21 = matrix1.M21 / matrix2.M21; result.M22 = matrix1.M22 / matrix2.M22; result.M23 = matrix1.M23 / matrix2.M23; result.M24 = matrix1.M24 / matrix2.M24;
            result.M31 = matrix1.M31 / matrix2.M31; result.M32 = matrix1.M32 / matrix2.M32; result.M33 = matrix1.M33 / matrix2.M33; result.M34 = matrix1.M34 / matrix2.M34;
            result.M41 = matrix1.M41 / matrix2.M41; result.M42 = matrix1.M42 / matrix2.M42; result.M43 = matrix1.M43 / matrix2.M43; result.M44 = matrix1.M44 / matrix2.M44;

            return result;
        }

        public static void Divide(ref Matrix matrix1, ref Matrix matrix2, out Matrix result)
        {
            result.M11 = matrix1.M11 / matrix2.M11; result.M12 = matrix1.M12 / matrix2.M12; result.M13 = matrix1.M13 / matrix2.M13; result.M14 = matrix1.M14 / matrix2.M14;
            result.M21 = matrix1.M21 / matrix2.M21; result.M22 = matrix1.M22 / matrix2.M22; result.M23 = matrix1.M23 / matrix2.M23; result.M24 = matrix1.M24 / matrix2.M24;
            result.M31 = matrix1.M31 / matrix2.M31; result.M32 = matrix1.M32 / matrix2.M32; result.M33 = matrix1.M33 / matrix2.M33; result.M34 = matrix1.M34 / matrix2.M34;
            result.M41 = matrix1.M41 / matrix2.M41; result.M42 = matrix1.M42 / matrix2.M42; result.M43 = matrix1.M43 / matrix2.M43; result.M44 = matrix1.M44 / matrix2.M44;
        }


        /// <summary>
        /// Divide individual elements of given matrix matrix1 by scalar vlaue matrix2.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="divider"></param>
        /// <returns></returns>
        public static Matrix Divide(Matrix matrix1, float divider)
        {
            Matrix result;

            float invDiv = 1.0f / divider;

            result.M11 = matrix1.M11 * invDiv; result.M12 = matrix1.M12 * invDiv; result.M13 = matrix1.M13 * invDiv; result.M14 = matrix1.M14 * invDiv;
            result.M21 = matrix1.M21 * invDiv; result.M22 = matrix1.M22 * invDiv; result.M23 = matrix1.M23 * invDiv; result.M24 = matrix1.M24 * invDiv;
            result.M31 = matrix1.M31 * invDiv; result.M32 = matrix1.M32 * invDiv; result.M33 = matrix1.M33 * invDiv; result.M34 = matrix1.M34 * invDiv;
            result.M41 = matrix1.M41 * invDiv; result.M42 = matrix1.M42 * invDiv; result.M43 = matrix1.M43 * invDiv; result.M44 = matrix1.M44 * invDiv;

            return result;
        }

        public static void Divide(ref Matrix matrix1, float divider, out Matrix result)
        {
            float invDiv = 1.0f / divider;

            result.M11 = matrix1.M11 * invDiv; result.M12 = matrix1.M12 * invDiv; result.M13 = matrix1.M13 * invDiv; result.M14 = matrix1.M14 * invDiv;
            result.M21 = matrix1.M21 * invDiv; result.M22 = matrix1.M22 * invDiv; result.M23 = matrix1.M23 * invDiv; result.M24 = matrix1.M24 * invDiv;
            result.M31 = matrix1.M31 * invDiv; result.M32 = matrix1.M32 * invDiv; result.M33 = matrix1.M33 * invDiv; result.M34 = matrix1.M34 * invDiv;
            result.M41 = matrix1.M41 * invDiv; result.M42 = matrix1.M42 * invDiv; result.M43 = matrix1.M43 * invDiv; result.M44 = matrix1.M44 * invDiv;
        }
        #endregion

        #region Operator overloads
        // Unary operator overloads
        public static Matrix operator -(Matrix matrix1)
        {
            Matrix m;

            m.M11 = -matrix1.M11; m.M12 = -matrix1.M12; m.M13 = -matrix1.M13; m.M14 = -matrix1.M14;
            m.M21 = -matrix1.M21; m.M22 = -matrix1.M22; m.M23 = -matrix1.M23; m.M24 = -matrix1.M24;
            m.M31 = -matrix1.M31; m.M32 = -matrix1.M32; m.M33 = -matrix1.M33; m.M34 = -matrix1.M34;
            m.M41 = -matrix1.M41; m.M42 = -matrix1.M42; m.M43 = -matrix1.M43; m.M44 = -matrix1.M44;

            return m;
        }

        // Binary operator overloads
        public static bool operator == (Matrix matrix1, Matrix matrix2)
        {
            return (
                matrix1.M11 == matrix2.M11 && matrix1.M22 == matrix2.M22 && matrix1.M33 == matrix2.M33 && matrix1.M44 == matrix2.M44 && // Check Diagnal element first for early out.
                                              matrix1.M12 == matrix2.M12 && matrix1.M13 == matrix2.M13 && matrix1.M14 == matrix2.M14 &&
                matrix1.M21 == matrix2.M21                               && matrix1.M23 == matrix2.M23 && matrix1.M24 == matrix2.M24 &&
                matrix1.M31 == matrix2.M31 && matrix1.M32 == matrix2.M32                               && matrix1.M34 == matrix2.M34 &&
                matrix1.M41 == matrix2.M41 && matrix1.M42 == matrix2.M42 && matrix1.M43 == matrix2.M43                                   );
        }

        public static bool operator !=(Matrix matrix1, Matrix matrix2)
        {
            return (
                matrix1.M11 != matrix2.M11 || matrix1.M12 != matrix2.M12 || matrix1.M13 != matrix2.M13 || matrix1.M14 != matrix2.M14 ||
                matrix1.M21 != matrix2.M21 || matrix1.M22 != matrix2.M22 || matrix1.M23 != matrix2.M23 || matrix1.M24 != matrix2.M24 ||
                matrix1.M31 != matrix2.M31 || matrix1.M32 != matrix2.M32 || matrix1.M33 != matrix2.M33 || matrix1.M34 != matrix2.M34 ||
                matrix1.M41 != matrix2.M41 || matrix1.M42 != matrix2.M42 || matrix1.M43 != matrix2.M43 || matrix1.M44 != matrix2.M44 );
        }

        public static Matrix operator +(Matrix matrix1, Matrix matrix2)
        {
            Matrix m;

            m.M11 = matrix1.M11 + matrix2.M11; m.M12 = matrix1.M12 + matrix2.M12; m.M13 = matrix1.M13 + matrix2.M13; m.M14 = matrix1.M14 + matrix2.M14;
            m.M21 = matrix1.M21 + matrix2.M21; m.M22 = matrix1.M22 + matrix2.M22; m.M23 = matrix1.M23 + matrix2.M23; m.M24 = matrix1.M24 + matrix2.M24;
            m.M31 = matrix1.M31 + matrix2.M31; m.M32 = matrix1.M32 + matrix2.M32; m.M33 = matrix1.M33 + matrix2.M33; m.M34 = matrix1.M34 + matrix2.M34;
            m.M41 = matrix1.M41 + matrix2.M41; m.M42 = matrix1.M42 + matrix2.M42; m.M43 = matrix1.M43 + matrix2.M43; m.M44 = matrix1.M44 + matrix2.M44;

            return m;
        }

        public static Matrix operator -(Matrix matrix1, Matrix matrix2)
        {
            Matrix m;

            m.M11 = matrix1.M11 - matrix2.M11; m.M12 = matrix1.M12 - matrix2.M12; m.M13 = matrix1.M13 - matrix2.M13; m.M14 = matrix1.M14 - matrix2.M14;
            m.M21 = matrix1.M21 - matrix2.M21; m.M22 = matrix1.M22 - matrix2.M22; m.M23 = matrix1.M23 - matrix2.M23; m.M24 = matrix1.M24 - matrix2.M24;
            m.M31 = matrix1.M31 - matrix2.M31; m.M32 = matrix1.M32 - matrix2.M32; m.M33 = matrix1.M33 - matrix2.M33; m.M34 = matrix1.M34 - matrix2.M34;
            m.M41 = matrix1.M41 - matrix2.M41; m.M42 = matrix1.M42 - matrix2.M42; m.M43 = matrix1.M43 - matrix2.M43; m.M44 = matrix1.M44 - matrix2.M44;

            return m;
        }

		public static Matrix operator *(Matrix matrix1, Windows.Devices.Sensors.SensorRotationMatrix sensorMatrix)
		{
			return matrix1 * new Matrix(sensorMatrix.M11, sensorMatrix.M12, sensorMatrix.M13, 0,
							  sensorMatrix.M21, sensorMatrix.M22, sensorMatrix.M23, 0,
							  sensorMatrix.M31, sensorMatrix.M32, sensorMatrix.M33, 0,
							  0, 0, 0, 1);
		}

        public static Matrix operator *(Matrix matrix1, Matrix matrix2)
        {
            Matrix m;

            // First row
            m.M11 = matrix1.M11 * matrix2.M11 + matrix1.M12 * matrix2.M21 + matrix1.M13 * matrix2.M31 + matrix1.M14 * matrix2.M41;
            m.M12 = matrix1.M11 * matrix2.M12 + matrix1.M12 * matrix2.M22 + matrix1.M13 * matrix2.M32 + matrix1.M14 * matrix2.M42;
            m.M13 = matrix1.M11 * matrix2.M13 + matrix1.M12 * matrix2.M23 + matrix1.M13 * matrix2.M33 + matrix1.M14 * matrix2.M43;
            m.M14 = matrix1.M11 * matrix2.M14 + matrix1.M12 * matrix2.M24 + matrix1.M13 * matrix2.M34 + matrix1.M14 * matrix2.M44;
            // Second row
            m.M21 = matrix1.M21 * matrix2.M11 + matrix1.M22 * matrix2.M21 + matrix1.M23 * matrix2.M31 + matrix1.M24 * matrix2.M41;
            m.M22 = matrix1.M21 * matrix2.M12 + matrix1.M22 * matrix2.M22 + matrix1.M23 * matrix2.M32 + matrix1.M24 * matrix2.M42;
            m.M23 = matrix1.M21 * matrix2.M13 + matrix1.M22 * matrix2.M23 + matrix1.M23 * matrix2.M33 + matrix1.M24 * matrix2.M43;
            m.M24 = matrix1.M21 * matrix2.M14 + matrix1.M22 * matrix2.M24 + matrix1.M23 * matrix2.M34 + matrix1.M24 * matrix2.M44;
            // Third row
            m.M31 = matrix1.M31 * matrix2.M11 + matrix1.M32 * matrix2.M21 + matrix1.M33 * matrix2.M31 + matrix1.M34 * matrix2.M41;
            m.M32 = matrix1.M31 * matrix2.M12 + matrix1.M32 * matrix2.M22 + matrix1.M33 * matrix2.M32 + matrix1.M34 * matrix2.M42;
            m.M33 = matrix1.M31 * matrix2.M13 + matrix1.M32 * matrix2.M23 + matrix1.M33 * matrix2.M33 + matrix1.M34 * matrix2.M43;
            m.M34 = matrix1.M31 * matrix2.M14 + matrix1.M32 * matrix2.M24 + matrix1.M33 * matrix2.M34 + matrix1.M34 * matrix2.M44;
            // Forth row
            m.M41 = matrix1.M41 * matrix2.M11 + matrix1.M42 * matrix2.M21 + matrix1.M43 * matrix2.M31 + matrix1.M44 * matrix2.M41;
            m.M42 = matrix1.M41 * matrix2.M12 + matrix1.M42 * matrix2.M22 + matrix1.M43 * matrix2.M32 + matrix1.M44 * matrix2.M42;
            m.M43 = matrix1.M41 * matrix2.M13 + matrix1.M42 * matrix2.M23 + matrix1.M43 * matrix2.M33 + matrix1.M44 * matrix2.M43;
            m.M44 = matrix1.M41 * matrix2.M14 + matrix1.M42 * matrix2.M24 + matrix1.M43 * matrix2.M34 + matrix1.M44 * matrix2.M44;

            return m;
        }

        public static Matrix operator *(Matrix matrix, float scaleFactor)
        {
            Matrix m;

            float s = scaleFactor;
            m.M11 = matrix.M11 * s; m.M12 = matrix.M12 * s; m.M13 = matrix.M13 * s; m.M14 = matrix.M14 * s;
            m.M21 = matrix.M21 * s; m.M22 = matrix.M22 * s; m.M23 = matrix.M23 * s; m.M24 = matrix.M24 * s;
            m.M31 = matrix.M31 * s; m.M32 = matrix.M32 * s; m.M33 = matrix.M33 * s; m.M34 = matrix.M34 * s;
            m.M41 = matrix.M41 * s; m.M42 = matrix.M42 * s; m.M43 = matrix.M43 * s; m.M44 = matrix.M44 * s;

            return m;
        }

        public static Matrix operator *(float scaleFactor, Matrix matrix)
        {
            Matrix m;

            float s = scaleFactor;
            m.M11 = matrix.M11 * s; m.M12 = matrix.M12 * s; m.M13 = matrix.M13 * s; m.M14 = matrix.M14 * s;
            m.M21 = matrix.M21 * s; m.M22 = matrix.M22 * s; m.M23 = matrix.M23 * s; m.M24 = matrix.M24 * s;
            m.M31 = matrix.M31 * s; m.M32 = matrix.M32 * s; m.M33 = matrix.M33 * s; m.M34 = matrix.M34 * s;
            m.M41 = matrix.M41 * s; m.M42 = matrix.M42 * s; m.M43 = matrix.M43 * s; m.M44 = matrix.M44 * s;

            return m;
        }

        public static Matrix operator /(Matrix matrix1, Matrix matrix2)
        {
            Matrix m;

            m.M11 = matrix1.M11 / matrix2.M11; m.M12 = matrix1.M12 / matrix2.M12; m.M13 = matrix1.M13 / matrix2.M13; m.M14 = matrix1.M14 / matrix2.M14;
            m.M21 = matrix1.M21 / matrix2.M21; m.M22 = matrix1.M22 / matrix2.M22; m.M23 = matrix1.M23 / matrix2.M23; m.M24 = matrix1.M24 / matrix2.M24;
            m.M31 = matrix1.M31 / matrix2.M31; m.M32 = matrix1.M32 / matrix2.M32; m.M33 = matrix1.M33 / matrix2.M33; m.M34 = matrix1.M34 / matrix2.M34;
            m.M41 = matrix1.M41 / matrix2.M41; m.M42 = matrix1.M42 / matrix2.M42; m.M43 = matrix1.M43 / matrix2.M43; m.M44 = matrix1.M44 / matrix2.M44;

            return m;
        }

        public static Matrix operator /(Matrix matrix1, float divider)
        {
            Matrix m;

            float invDiv = 1.0f / divider;

            m.M11 = matrix1.M11 * invDiv; m.M12 = matrix1.M12 * invDiv; m.M13 = matrix1.M13 * invDiv; m.M14 = matrix1.M14 * invDiv;
            m.M21 = matrix1.M21 * invDiv; m.M22 = matrix1.M22 * invDiv; m.M23 = matrix1.M23 * invDiv; m.M24 = matrix1.M24 * invDiv;
            m.M31 = matrix1.M31 * invDiv; m.M32 = matrix1.M32 * invDiv; m.M33 = matrix1.M33 * invDiv; m.M34 = matrix1.M34 * invDiv;
            m.M41 = matrix1.M41 * invDiv; m.M42 = matrix1.M42 * invDiv; m.M43 = matrix1.M43 * invDiv; m.M44 = matrix1.M44 * invDiv;

            return m;
        }
        #endregion

    }
}
