﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Microsoft.Xna.Framework.Graphics
{
	public struct Viewport
	{
		private int _x;
		private int _y;
		private int _width;
		private int _height;
		private float _minZ;
		private float _maxZ;

		public Viewport(int x, int y, int width, int height)
		{
			this._x = x;
			this._y = y;
			this._width = width;
			this._height = height;
			this._minZ = 0f;
			this._maxZ = 1f;
		}

		public int X
		{
			get
			{
				return this._x;
			}
			set
			{
				this._x = value;
			}
		}
		public int Y
		{
			get
			{
				return this._y;
			}
			set
			{
				this._y = value;
			}
		}
		public int Width
		{
			get
			{
				return this._width;
			}
			set
			{
				this._width = value;
			}
		}
		public int Height
		{
			get
			{
				return this._height;
			}
			set
			{
				this._height = value;
			}
		}
		public float MinDepth
		{
			get
			{
				return this._minZ;
			}
			set
			{
				this._minZ = value;
			}
		}
		public float MaxDepth
		{
			get
			{
				return this._maxZ;
			}
			set
			{
				this._maxZ = value;
			}
		}
		private static bool WithinEpsilon(float a, float b)
		{
			float num = a - b;
			return ((-1.401298E-45f <= num) && (num <= float.Epsilon));
		}

		public Vector3 Project(Vector3 source, Matrix projection, Matrix view, Matrix world)
		{
			Matrix matrix = Matrix.Multiply(Matrix.Multiply(world, view), projection);
			Vector3 vector = Vector3.Transform(source, matrix);
			float a = (((source.X * matrix.M14) + (source.Y * matrix.M24)) + (source.Z * matrix.M34)) + matrix.M44;
			if (!WithinEpsilon(a, 1f))
			{
				vector = (Vector3)(vector / a);
			}
			vector.X = (((vector.X + 1f) * 0.5f) * this.Width) + this.X;
			vector.Y = (((-vector.Y + 1f) * 0.5f) * this.Height) + this.Y;
			vector.Z = (vector.Z * (this.MaxDepth - this.MinDepth)) + this.MinDepth;
			return vector;
		}

		public Vector3 Unproject(Vector3 source, Matrix projection, Matrix view, Matrix world)
		{
			Matrix matrix = Matrix.Invert(Matrix.Multiply(Matrix.Multiply(world, view), projection));
			source.X = (((source.X - this.X) / ((float)this.Width)) * 2f) - 1f;
			source.Y = -((((source.Y - this.Y) / ((float)this.Height)) * 2f) - 1f);
			source.Z = (source.Z - this.MinDepth) / (this.MaxDepth - this.MinDepth);
			Vector3 vector = Vector3.Transform(source, matrix);
			float a = (((source.X * matrix.M14) + (source.Y * matrix.M24)) + (source.Z * matrix.M34)) + matrix.M44;
			if (!WithinEpsilon(a, 1f))
			{
				vector = (Vector3)(vector / a);
			}
			return vector;
		}
		public float AspectRatio
		{
			get
			{
				if ((this._height != 0) && (this._width != 0))
				{
					return (((float)this._width) / ((float)this._height));
				}
				return 0f;
			}
		}
	}
}
